#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

# import os.path as op

logger = logging.getLogger(__name__)
import numpy as np
import scipy
import scipy.ndimage


def resize_to_shape(data, shape, zoom=None, mode="nearest", order=0):
    """
    Function resize input data to specific shape.
    :param data: input 3d array-like data
    :param shape: shape of output data
    :param zoom: zoom is used for back compatibility
    :mode: default is 'nearest'
    """
    # @TODO remove old code in except part
    # TODO use function from library in future

    try:
        # rint 'pred vyjimkou'
        # aise Exception ('test without skimage')
        # rint 'za vyjimkou'
        import skimage
        import skimage.transform

        # Now we need reshape  seeds and segmentation to original size

        # with warnings.catch_warnings():
        #     warnings.filterwarnings("ignore", ".*'constant', will be changed to.*")
        segm_orig_scale = skimage.transform.resize(
            data, shape, order=0, preserve_range=True, mode="reflect"
        )

        segmentation = segm_orig_scale
        logger.debug("resize to orig with skimage")
    except:
        if zoom is None:
            zoom = shape / np.asarray(data.shape).astype(np.double)
        segmentation = resize_to_shape_with_zoom(
            data, zoom=zoom, mode=mode, order=order
        )

    return segmentation


def resize_to_shape_with_zoom(data, shape, zoom, mode="nearest", order=0):
    import scipy
    import scipy.ndimage

    dtype = data.dtype

    segm_orig_scale = scipy.ndimage.zoom(
        data, 1.0 / zoom, mode=mode, order=order
    ).astype(dtype)
    logger.debug("resize to orig with scipy.ndimage")

    # @TODO odstranit hack pro oříznutí na stejnou velikost
    # v podstatě je to vyřešeno, ale nechalo by se to dělat elegantněji v zoom
    # tam je bohužel patrně bug
    # rint 'd3d ', self.data3d.shape
    # rint 's orig scale shape ', segm_orig_scale.shape
    shp = [
        np.min([segm_orig_scale.shape[0], shape[0]]),
        np.min([segm_orig_scale.shape[1], shape[1]]),
        np.min([segm_orig_scale.shape[2], shape[2]]),
    ]
    # elf.data3d = self.data3d[0:shp[0], 0:shp[1], 0:shp[2]]
    # mport ipdb; ipdb.set_trace() # BREAKPOINT

    segmentation = np.zeros(shape, dtype=dtype)
    segmentation[0 : shp[0], 0 : shp[1], 0 : shp[2]] = segm_orig_scale[
        0 : shp[0], 0 : shp[1], 0 : shp[2]
    ]

    del segm_orig_scale
    return segmentation


def seed_zoom(seeds, zoom):
    """
    Smart zoom for sparse matrix. If there is resize to bigger resolution
    thin line of label could be lost. This function prefers labels larger
    then zero. If there is only one small voxel in larger volume with zeros
    it is selected.
    """
    # import scipy
    # loseeds=seeds
    labels = np.unique(seeds)
    # remove first label - 0
    labels = np.delete(labels, 0)
    # @TODO smart interpolation for seeds in one block
    #        loseeds = scipy.ndimage.interpolation.zoom(
    #            seeds, zoom, order=0)
    loshape = np.ceil(np.array(seeds.shape) * 1.0 / zoom).astype(np.int)
    loseeds = np.zeros(loshape, dtype=np.int8)
    loseeds = loseeds.astype(np.int8)
    for label in labels:
        a, b, c = np.where(seeds == label)
        loa = np.round(a // zoom)
        lob = np.round(b // zoom)
        loc = np.round(c // zoom)
        # loseeds = np.zeros(loshape)

        loseeds[loa, lob, loc] += label
        # this is to detect conflict seeds
        loseeds[loseeds > label] = 100

    # remove conflict seeds
    loseeds[loseeds > 99] = 0

    # import py3DSeedEditor
    # ped = py3DSeedEditor.py3DSeedEditor(loseeds)
    # ped.show()

    return loseeds


def zoom_to_shape(data, shape, dtype=None):
    """
    Zoom data to specific shape.
    """
    import scipy
    import scipy.ndimage

    zoomd = np.array(shape) / np.array(data.shape, dtype=np.double)
    import warnings

    datares = scipy.ndimage.interpolation.zoom(data, zoomd, order=0, mode="reflect")

    if datares.shape != shape:
        logger.warning("Zoom with different output shape")
    dataout = np.zeros(shape, dtype=dtype)
    shpmin = np.minimum(dataout.shape, shape)

    dataout[: shpmin[0], : shpmin[1], : shpmin[2]] = datares[
        : shpmin[0], : shpmin[1], : shpmin[2]
    ]
    return datares


def select_objects_by_seeds(
    binar_data, seeds, ignore_background_seeds=True, background_label=0
):

    labeled_data, length = scipy.ndimage.label(binar_data)
    selected_labels = list(np.unique(labeled_data[seeds > 0]))
    # selected_labels.pop(0)
    # pop the background label
    output = np.zeros_like(binar_data)
    for label in selected_labels:
        selection = labeled_data == label
        # copy from input image to output. If there will be seeds in background, the 0 is copied
        if ignore_background_seeds and (binar_data[selection][0] == background_label):
            pass
        else:
            # output[selection] = binar_data[selection]
            output[selection] = 1
    # import sed3
    # ed =sed3.sed3(labeled_data, contour=output, seeds=seeds)
    # ed.show()
    return output


# def getPriorityObjects(*args, **kwargs):
#     logger.warning("Function getPriorityObjects has been renamed. Use get_priority_objects().")
#     DeprecationWarning("Function getPriorityObjects has been renamed. Use get_priority_objects().")
#     return get_priority_objects(*args, **kwargs)
#
# def get_priority_objects(data, nObj=1, seeds=None, debug=False):
#     """
#     Get N biggest objects from the selection or the object with seed.
#
#     :param data:  labeled ndarray
#     :param nObj:  number of objects
#     :param seeds: ndarray. Objects on non zero positions are returned
#     :param debug: bool.
#     :return: binar image with selected objects
#     """
#
#     # Oznaceni dat.
#     # labels - oznacena data.
#     # length - pocet rozdilnych oznaceni.
#     dataLabels, length = scipy.ndimage.label(data)
#
#     logger.info('Olabelovano oblasti: ' + str(length))
#
#     if debug:
#         logger.debug('data labels: ' + str(dataLabels))
#
#     # Uzivatel si nevybral specificke objekty.
#     if (seeds == None):
#
#         logger.info('Vraceni bez seedu')
#         logger.debug('Objekty: ' + str(nObj))
#
#         # Zjisteni nejvetsich objektu.
#         arrayLabelsSum, arrayLabels = areaIndexes(dataLabels, length)
#         # Serazeni labelu podle velikosti oznacenych dat (prvku / ploch).
#         arrayLabelsSum, arrayLabels = selectSort(arrayLabelsSum, arrayLabels)
#
#         returning = None
#         label = 0
#         stop = nObj - 1
#
#         # Budeme postupne prochazet arrayLabels a postupne pridavat jednu
#         # oblast za druhou (od te nejvetsi - mimo nuloveho pozadi) dokud
#         # nebudeme mit dany pocet objektu (nObj).
#         while label <= stop:
#
#             if label >= len(arrayLabels):
#                 break
#
#             if arrayLabels[label] != 0:
#                 if returning == None:
#                     # "Prvni" iterace
#                     returning = data * (dataLabels == arrayLabels[label])
#                 else:
#                     # Jakakoli dalsi iterace
#                     returning = returning + data * \
#                                 (dataLabels == arrayLabels[label])
#             else:
#                 # Musime prodlouzit hledany interval, protoze jsme narazili na
#                 # nulove pozadi.
#                 stop = stop + 1
#
#             label = label + 1
#
#             if debug:
#                 logger.debug(str(label - 1) + ': ' + str(returning))
#
#         if returning == None:
#             logger.info(
#                 'Zadna validni olabelovana data! (DEBUG: returning == None)')
#
#         return returning
#
#     # Uzivatel si vybral specificke objekty (seeds != None).
#     else:
#
#         logger.info('Vraceni se seedy')
#
#         # Zalozeni pole pro ulozeni seedu
#         arrSeed = []
#         # Zjisteni poctu seedu.
#         stop = seeds[0].size
#         tmpSeed = 0
#         dim = np.ndim(dataLabels)
#         for index in range(0, stop):
#             # Tady se ukladaji labely na mistech, ve kterych kliknul uzivatel.
#             if dim == 3:
#                 # 3D data.
#                 tmpSeed = dataLabels[
#                     seeds[0][index], seeds[1][index], seeds[2][index]]
#             elif dim == 2:
#                 # 2D data.
#                 tmpSeed = dataLabels[seeds[0][index], seeds[1][index]]
#
#             # Tady opet pocitam s tim, ze oznaceni nulou pripada cerne oblasti
#             # (pozadi).
#             if tmpSeed != 0:
#                 # Pokud se nejedna o pozadi (cernou oblast), tak se novy seed
#                 # ulozi do pole "arrSeed"
#                 arrSeed.append(tmpSeed)
#
#         # Pokud existuji vhodne labely, vytvori se nova data k vraceni.
#         # Pokud ne, vrati se "None" typ. { Deprecated: Pokud ne, vrati se cela
#         # nafiltrovana data, ktera do funkce prisla (nedojde k vraceni
#         # specifickych objektu). }
#         if len(arrSeed) > 0:
#
#             # Zbaveni se duplikatu.
#             arrSeed = list(set(arrSeed))
#             if debug:
#                 logger.debug('seed list:' + str(arrSeed))
#
#             logger.info(
#                 'Ruznych prioritnich objektu k vraceni: ' +
#                 str(len(arrSeed))
#             )
#
#             # Vytvoreni vystupu - postupne pricitani dat prislunych specif.
#             # labelu.
#             returning = None
#             for index in range(0, len(arrSeed)):
#
#                 if returning == None:
#                     returning = data * (dataLabels == arrSeed[index])
#                 else:
#                     returning = returning + data * \
#                                 (dataLabels == arrSeed[index])
#
#                 if debug:
#                     logger.debug((str(index)) + ':' + str(returning))
#
#             return returning
#
#         else:
#
#             logger.info(
#                 'Zadna validni data k vraceni - zadne prioritni objekty ' +
#                 'nenalezeny (DEBUG: function getPriorityObjects:' +
#                 str(len(arrSeed) == 0))
#             return None
#
# def areaIndexes(labels, num):
#     """
#
#     Zjisti cetnosti jednotlivych oznacenych ploch (labeled areas)
#         input:
#             labels - data s aplikovanymi oznacenimi
#             num - pocet pouzitych oznaceni
#
#         returns:
#             dve pole - prvni sumy, druhe indexy
#
#     """
#
#     arrayLabelsSum = []
#     arrayLabels = []
#     for index in range(0, num + 1):
#         arrayLabels.append(index)
#         sumOfLabel = numpy.sum(labels == index)
#         arrayLabelsSum.append(sumOfLabel)
#
#     return arrayLabelsSum, arrayLabels
#
#
# def selectSort(list1, list2):
#     """
#     Razeni 2 poli najednou (list) pomoci metody select sort
#         input:
#             list1 - prvni pole (hlavni pole pro razeni)
#             list2 - druhe pole (vedlejsi pole) (kopirujici pozice pro razeni
#                 podle hlavniho pole list1)
#
#         returns:
#             dve serazena pole - hodnoty se ridi podle prvniho pole, druhe
#                 "kopiruje" razeni
#     """
#
#     length = len(list1)
#     for index in range(0, length):
#         min = index
#         for index2 in range(index + 1, length):
#             if list1[index2] > list1[min]:
#                 min = index2
#         # Prohozeni hodnot hlavniho pole
#         list1[index], list1[min] = list1[min], list1[index]
#         # Prohozeni hodnot vedlejsiho pole
#         list2[index], list2[min] = list2[min], list2[index]
#
#     return list1, list2


def crop(data, crinfo):
    """
    Crop the data.

    crop(data, crinfo)

    :param crinfo: min and max for each axis - [[minX, maxX], [minY, maxY], [minZ, maxZ]]

    """
    crinfo = fix_crinfo(crinfo)
    return data[
        __int_or_none(crinfo[0][0]) : __int_or_none(crinfo[0][1]),
        __int_or_none(crinfo[1][0]) : __int_or_none(crinfo[1][1]),
        __int_or_none(crinfo[2][0]) : __int_or_none(crinfo[2][1]),
    ]


def __int_or_none(number):
    if number is not None:
        number = int(number)
    return number


def combinecrinfo(crinfo1, crinfo2):
    """
    Combine two crinfos. First used is crinfo1, second used is crinfo2.
    """
    crinfo1 = fix_crinfo(crinfo1)
    crinfo2 = fix_crinfo(crinfo2)

    crinfo = [
        [crinfo1[0][0] + crinfo2[0][0], crinfo1[0][0] + crinfo2[0][1]],
        [crinfo1[1][0] + crinfo2[1][0], crinfo1[1][0] + crinfo2[1][1]],
        [crinfo1[2][0] + crinfo2[2][0], crinfo1[2][0] + crinfo2[2][1]],
    ]

    return crinfo


def crinfo_from_specific_data(data, margin=0):
    """
    Create crinfo of minimum orthogonal nonzero block in input data.

    :param data: input data
    :param margin: add margin to minimum block
    :return:
    """
    # hledáme automatický ořez, nonzero dá indexy
    logger.debug("crinfo")
    logger.debug(str(margin))
    nzi = np.nonzero(data)
    logger.debug(str(nzi))

    if np.isscalar(margin):
        margin = [margin] * 3

    x1 = np.min(nzi[0]) - margin[0]
    x2 = np.max(nzi[0]) + margin[0] + 1
    y1 = np.min(nzi[1]) - margin[0]
    y2 = np.max(nzi[1]) + margin[0] + 1
    z1 = np.min(nzi[2]) - margin[0]
    z2 = np.max(nzi[2]) + margin[0] + 1

    # ošetření mezí polí
    if x1 < 0:
        x1 = 0
    if y1 < 0:
        y1 = 0
    if z1 < 0:
        z1 = 0

    if x2 > data.shape[0]:
        x2 = data.shape[0] - 1
    if y2 > data.shape[1]:
        y2 = data.shape[1] - 1
    if z2 > data.shape[2]:
        z2 = data.shape[2] - 1

    # ořez
    crinfo = [[x1, x2], [y1, y2], [z1, z2]]
    return crinfo


def uncrop(data, crinfo, orig_shape, resize=False, outside_mode="constant", cval=0):
    """
    Put some boundary to input image.


    :param data: input data
    :param crinfo: array with minimum and maximum index along each axis
        [[minX, maxX],[minY, maxY],[minZ, maxZ]]. If crinfo is None, the whole input image is placed into [0, 0, 0].
        If crinfo is just series of three numbers, it is used as an initial point for input image placement.
    :param orig_shape: shape of uncropped image
    :param resize: True or False (default). Usefull if the data.shape does not fit to crinfo shape.
    :param outside_mode: 'constant', 'nearest'
    :return:
    """

    if crinfo is None:
        crinfo = list(zip([0] * data.ndim, orig_shape))
    elif np.asarray(crinfo).size == data.ndim:
        crinfo = list(zip(crinfo, np.asarray(crinfo) + data.shape))

    crinfo = fix_crinfo(crinfo)
    data_out = np.ones(orig_shape, dtype=data.dtype) * cval

    # print 'uncrop ', crinfo
    # print orig_shape
    # print data.shape
    if resize:
        data = resize_to_shape(data, crinfo[:, 1] - crinfo[:, 0])

    startx = np.round(crinfo[0][0]).astype(int)
    starty = np.round(crinfo[1][0]).astype(int)
    startz = np.round(crinfo[2][0]).astype(int)

    data_out[
        # np.round(crinfo[0][0]).astype(int):np.round(crinfo[0][1]).astype(int)+1,
        # np.round(crinfo[1][0]).astype(int):np.round(crinfo[1][1]).astype(int)+1,
        # np.round(crinfo[2][0]).astype(int):np.round(crinfo[2][1]).astype(int)+1
        startx : startx + data.shape[0],
        starty : starty + data.shape[1],
        startz : startz + data.shape[2],
    ] = data

    if outside_mode == "nearest":
        # for ax in range(data.ndims):
        # ax = 0

        # copy border slice to pixels out of boundary - the higher part
        for ax in range(data.ndim):
            # the part under the crop
            start = np.round(crinfo[ax][0]).astype(int)
            slices = [slice(None), slice(None), slice(None)]
            slices[ax] = start
            repeated_slice = np.expand_dims(data_out[slices], ax)
            append_sz = start
            if append_sz > 0:
                tile0 = np.repeat(repeated_slice, append_sz, axis=ax)
                slices = [slice(None), slice(None), slice(None)]
                slices[ax] = slice(None, start)
                # data_out[start + data.shape[ax] : , :, :] = tile0
                data_out[slices] = tile0
                # plt.imshow(np.squeeze(repeated_slice))
                # plt.show()

            # the part over the crop
            start = np.round(crinfo[ax][0]).astype(int)
            slices = [slice(None), slice(None), slice(None)]
            slices[ax] = start + data.shape[ax] - 1
            repeated_slice = np.expand_dims(data_out[slices], ax)
            append_sz = data_out.shape[ax] - (start + data.shape[ax])
            if append_sz > 0:
                tile0 = np.repeat(repeated_slice, append_sz, axis=ax)
                slices = [slice(None), slice(None), slice(None)]
                slices[ax] = slice(start + data.shape[ax], None)
                # data_out[start + data.shape[ax] : , :, :] = tile0
                data_out[slices] = tile0
                # plt.imshow(np.squeeze(repeated_slice))
                # plt.show()

    return data_out


def fix_crinfo(crinfo, to="axis"):
    """
    Function recognize order of crinfo and convert it to proper format.
    """

    crinfo = np.asarray(crinfo)
    if crinfo.shape[0] == 2:
        crinfo = crinfo.T

    return crinfo



#! /usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import print_function

# import funkcí z jiného adresáře
import sys
import os.path
import unittest
import scipy
import numpy as np

import logging

logger = logging.getLogger(__name__)

path_to_script = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(path_to_script, "../src/"))

from nose.plugins.attrib import attr
from imcut import pycut


def fv_function(data, voxelsize, seeds=None, cls=None):
    """
    Creates feature vector for only data or for data from classes
    """

    fv1 = data.reshape(-1, 1)

    data2 = scipy.ndimage.filters.gaussian_filter(data, sigma=0.1)
    fv2 = data2.reshape(-1, 1)

    fv = np.hstack([fv1, fv2])

    if seeds is not None:
        logger.debug("seeds " + str(seeds))
        # print("seeds ", seeds)
        sd = seeds.reshape(-1, 1)
        selection = np.in1d(sd, cls)
        fv = fv[selection]
        sd = sd[selection]
        # sd = sd[]
        return fv, sd
    return fv


def box_data(noise_sigma=3):
    # data
    img3d = np.random.rand(32, 64, 64) * noise_sigma
    img3d[4:24, 12:32, 5:25] = img3d[4:24, 12:32, 5:25] + 30

    # seeds
    seeds = np.zeros([32, 64, 64], np.int8)
    seeds[9:12, 13:29, 18:25] = 1
    seeds[9:12, 4:9, 3:32] = 2
    # [mm]  10 x 10 x 10        # voxelsize_mm = [1, 4, 3]
    voxelsize_mm = [5, 5, 5]
    metadata = {"voxelsize_mm": voxelsize_mm}
    return img3d, seeds, voxelsize_mm


class PycutTest(unittest.TestCase):
    def setUp(self):
        self.segparams_ssgc = {
            # 'method':'graphcut',
            "method": "graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {
                "type": "gmmsame",
                "params": {"n_components": 2},
                # "return_only_objects_with_seeds": True,
                # 'fv_type': "fv_extern",
                # 'fv_extern': fv_function,
                # 'adaptation': 'original_data',
            },
        }
        self.segparams_ssgc = {
            # 'method':'graphcut',
            "method": "graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {
                "type": "gmmsame",
                "params": {"n_components": 2},
                # "return_only_objects_with_seeds": True,
                # 'fv_type': "fv_extern",
                # 'fv_extern': fv_function,
                # 'adaptation': 'original_data',
            },
        }
        self.segparams_lo2hi = {
            # 'method':'graphcut',
            "method": "graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {
                "type": "gmmsame",
                "params": {"n_components": 2}
                # 'fv_type': "fv_extern",
                # 'fv_extern': fv_function,
                # 'adaptation': 'original_data',
            },
        }
        self.segparams_hi2lo = {
            # 'method':'graphcut',
            "method": "graphcut_hi2lo",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {
                "type": "gmmsame",
                "params": {"n_components": 2}
                # 'fv_type': "fv_extern",
                # 'fv_extern': fv_function,
                # 'adaptation': 'original_data',
            },
        }

    @classmethod
    def setUpClass(cls):
        if sys.version_info.major < 3:
            cls.assertCountEqual = cls.assertItemsEqual

    @unittest.skipIf(
        os.environ.get("TRAVIS", True),
        "I dont know why this test fail on travis. Probably the RNG is allways the same",
    )
    def test_simple_graph_cut(self):
        img, seg, seeds = self.make_data(64, 20)
        gc = pycut.ImageGraphCut(img, segparams=self.segparams_ssgc)
        gc.set_seeds(seeds)

        gc.run()
        # import sed3
        # ed = sed3.sed3((gc.segmentation==0).astype(np.double), contour=seg)
        # ed.show()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertLess(err, 600)

    @attr("interactive")
    def test_simple_graph_cut_interactive(self):
        img, seg, seeds = self.make_data(64, 20)
        gc = pycut.ImageGraphCut(img, segparams=self.segparams_ssgc)
        gc.set_seeds(seeds)

        # gc.run()
        gc.interactivity()
        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertLess(err, 600)

    def test_simple_graph_cut_show_similarity(self):
        img, seg, seeds = self.make_data(64, 20)
        gc = pycut.ImageGraphCut(img, segparams=self.segparams_ssgc)
        gc.set_seeds(seeds)

        gc.run()
        gc.debug_show_reconstructed_similarity(show=False)
        gc.debug_show_model(start=-500, stop=500, show=False)
        # import sed3
        # ed = sed3.sed3((gc.segmentation==0).astype(np.double), contour=seg)
        # ed.show()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertLess(err, 600)

    def test_simple_graph_cut_feature_vector_fv001_intensity_and_blur(self):
        # test feature vector with
        segparams = {
            # 'method':'graphcut',
            "method": "graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {
                "type": "gmmsame",
                "params": {"n_components": 2},
                "fv_type": "intensity_and_blur"
                # 'fv_type': "fv_extern",
                # 'fv_extern': fv_function,
                # 'adaptation': 'original_data',
            },
        }
        img, seg, seeds = self.make_data(64, 20)
        gc = pycut.ImageGraphCut(img, segparams=segparams)
        gc.set_seeds(seeds)

        gc.run()
        # outputs looks strange
        # gc.show_similarity(show=True)
        # gc.show_model(start=-500, stop=500, show=True)
        # import sed3
        # ed = sed3.sed3(img, contour=gc.segmentation)
        # ed.show()
        # ed = sed3.sed3((gc.segmentation==0).astype(np.double), contour=seg)
        # ed.show()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        # TODO check multidimensional feature vector. It looks to be not sufficient.
        # self.assertLess(err, 2000)

    def test_gc_box_overfiting(self):
        data3d, seeds, voxelsize = box_data(noise_sigma=0.5)
        segparams = {
            # 'method':'graphcut',
            "method": "graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {
                "type": "gmmsame",
                "params": {"n_components": 3},
                # 'fv_type': "fv_extern",
                # 'fv_extern': fv_function,
                # 'adaptation': 'original_data',
            },
        }
        gc = pycut.ImageGraphCut(data3d, segparams=segparams, debug_images=False)
        gc.set_seeds(seeds)
        gc.run()
        # gc.interactivity()

        # import sed3
        # ed = sed3.sed3(data3d, contour=(gc.segmentation==0).astype(np.double) * 3)
        # ed.show()

    def test_simple_graph_cut_overfit_with_low_noise(self):
        img, seg, seeds = self.make_data(64, 20, sigma=20)
        segparams = {
            # 'method':'graphcut',
            "method": "graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {
                "type": "gmmsame",
                "params": {"n_components": 3},
                # 'fv_type': "fv_extern",
                # 'fv_extern': fv_function,
                # 'adaptation': 'original_data',
            },
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams, debug_images=False)
        gc.set_seeds(seeds)

        gc.run()
        # import sed3
        # ed = sed3.sed3(img, contour=(gc.segmentation==0).astype(np.double))
        # ed.show()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertLess(err, 600)

    @attr("interactive")
    def test_show_editor(self):
        """
        just run editor to see what is new
        Returns:

        """
        import imcut.seed_editor_qt
        import numpy as np
        from PyQt4.QtGui import QApplication

        app = QApplication(sys.argv)
        data = (np.random.rand(30, 31, 32) * 100).astype(np.int)
        data[15:40, 13:20, 10:18] += 50
        se = imcut.seed_editor_qt.QTSeedEditor(data)
        se.exec_()
        # self.assertTrue(False)

    # @attr('interactive')
    def test_draw_seed_function(self):
        """
        just run editor to see what is new
        Returns:

        """
        import imcut.seed_editor_qt
        import numpy as np
        from PyQt4.QtGui import QApplication

        app = QApplication(sys.argv)
        data = (np.random.rand(30, 31, 32) * 100).astype(np.int)
        data[15:40, 13:20, 10:18] += 50
        se = imcut.seed_editor_qt.QTSeedEditor(data)
        se.slice_box.seed_mark = 1  # left mouse button
        se.slice_box.last_position = [1, 3]
        se.slice_box.drawSeeds([10, 5])
        se.slice_box.seed_mark = 2  # left mouse button
        se.slice_box.last_position = [8, 1]
        se.slice_box.drawSeeds([7, 5])
        # se.exec_()

    # @TODO znovu zprovoznit test

    # @unittest.skip("Cekame, az to Tomas opravi")
    def make_data(self, sz=32, offset=0, sigma=80):
        seeds = np.zeros([sz, sz, sz], dtype=np.int8)
        seeds[offset + 12, offset + 9 : offset + 14, offset + 10] = 1
        seeds[offset + 20, offset + 18 : offset + 21, offset + 12] = 1
        img = np.ones([sz, sz, sz])
        img = img - seeds

        seeds[
            offset + 3 : offset + 15, offset + 2 : offset + 6, offset + 27 : offset + 29
        ] = 2
        img = scipy.ndimage.morphology.distance_transform_edt(img)
        segm = img < 7
        img = (100 * segm + sigma * np.random.random(img.shape)).astype(np.uint8)
        return img, segm, seeds

    def test_remove_repetitive(self):
        import time

        start = time.time()
        nlinks_not_unique = np.random.randint(0, 5, [100000, 3])
        nlinks = pycut.ms_remove_repetitive_link(nlinks_not_unique)
        elapsed = time.time() - start
        # print "elapsed ", elapsed

    def test_external_fv(self):
        """
        test external feature vector
        """

        img, seg, seeds = self.make_data(64, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {"fv_type": "fv_extern", "fv_extern": fv_function},
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams)
        gc.set_seeds(seeds)
        gc.run()
        # import sed3
        # ed = sed3.sed3(gc.segmentation==0, contour=seg)
        # ed.show()

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            600,
        )

    def test_test_fv_function(self):
        img, seg, seeds = self.make_data(64, 20)
        vs = [1, 1, 1]
        out = fv_function(img, voxelsize=vs)
        print(out.shape)
        out1, out2 = fv_function(img, vs, seeds, [1, 2])
        print(np.min(out1), np.max(out1), out1.shape)
        print(np.min(out2), np.max(out2), out2.shape)
        self.assertEqual(out.shape[0], np.prod(img.shape))
        self.assertEqual(out1.shape[0], out2.shape[0])
        self.assertEqual(np.min(out2), 1)
        self.assertEqual(np.max(out2), 2)

    def test_external_fv_with_save(self):
        """
        test external feature vector with save model in the middle of processing
        """

        img, seg, seeds = self.make_data(64, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {
                "type": "gmmsame",
                "fv_type": "fv_extern",
                "fv_extern": fv_function,
                "adaptation": "original_data",
            },
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams)
        gc.set_seeds(seeds)

        gc.run()
        # import sed3
        # ed = sed3.sed3((gc.segmentation==0).astype(np.double), contour=seg)
        # ed.show()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertLess(err, 600)

        mdl_stored_file = "test_model.p"
        gc.save(mdl_stored_file)

        # forget
        gc = None

        img, seg, seeds = self.make_data(56, 18)
        # there is only one change in mdl params
        segparams["modelparams"] = {
            "mdl_stored_file": mdl_stored_file,
            "fv_extern": fv_function,
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams)
        gc.set_seeds(seeds)
        gc.run()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertLess(err, 600)
        # import sed3
        # sed3.show_slices(img, contour=gc.segmentation==0, slice_step=6)

        # if we change the data there should be more error (assertMore)
        img = (img * 0.2).astype(np.uint8)
        # segparams['modelparams']['adaptation'] = 'original_data'
        # print(np.max(img))
        # print(np.min(img))
        gc = pycut.ImageGraphCut(img, segparams=segparams)
        gc.set_seeds(seeds)
        gc.run()

        m0 = gc.mdl.mdl[1]
        m1 = gc.mdl.mdl[2]
        logger.debug("model parameters")

        # import sed3
        # ed = sed3.sed3((gc.segmentation==0).astype(np.double), contour=seg)
        # ed.show()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertGreater(err, 600)
        # self.assertGreater(
        #     np.sum(
        #         np.abs(
        #             (gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        #     ),
        #     600)

        os.remove(mdl_stored_file)

    def test_apriori(self):
        """
        Test apriori segmentation. Make segmentation twice. First is used with gamma=0.2,
        second is used with gamma=0.9
        """

        img, seg, seeds = self.make_data(64, 20)
        apriori = np.zeros([64, 64, 64])

        apriori[:20, :20, :20] = 1
        segparams1 = {"apriori_gamma": 0.1}
        gc = pycut.ImageGraphCut(img, segparams=segparams1)
        gc.set_seeds(seeds)
        gc.apriori = apriori
        gc.run()
        # import sed3
        # ed = sed3.sed3(img, contour=(gc.segmentation==0))
        # ed.show()

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            600,
        )

        segparams2 = {"apriori_gamma": 0.9}
        gc = pycut.ImageGraphCut(img, segparams=segparams2)
        gc.set_seeds(seeds)
        gc.apriori = apriori
        gc.run()

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - apriori.astype(np.int8))
            ),
            600,
        )

    def test_multiscale_gc_seg(self):
        """
        Test multiscale segmentation
        """

        img, seg, seeds = self.make_data(64, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "multiscale_graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "block_size": 8,
            "tile_zoom_constant": 1,
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams)
        gc.set_seeds(seeds)
        gc.run()
        # import sed3
        # ed = sed3.sed3(gc.segmentation==0, contour=seg)
        # ed.show()

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            600,
        )

    def test_fast_multiscale_gc_seg(self):
        """
        Test fast multiscale segmentation
        """

        img, seg, seeds = self.make_data(64, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "multiscale_graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "block_size": 8,
            "tile_zoom_constant": 1,
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams)
        gc.set_seeds(seeds)
        gc.run()
        # import sed3
        # ed = sed3.sed3(gc.segmentation==0, contour=seg)
        # ed.show()

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            600,
        )

    # @unittest.skip("Cekame, az to Mire opravi")
    def test_ms_seg_compared_with_different_resolution(self):
        """
        Test multiscale segmentation compared with different resolution
        """

        img, seg, seeds = self.make_data(64, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "multiscale_graphcut_hi2lo",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "block_size": 8,
            "tile_zoom_constant": 1,
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams)
        gc.set_seeds(seeds)
        gc.run()
        # import sed3
        # ed = sed3.sed3(gc.segmentation==0, contour=seg)
        # ed.show()

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            600,
        )

        # different resolution
        # sz = [128,128,128]
        sz = [70, 70, 70]
        sz = [90, 90, 90]
        sz = [100, 100, 100]
        sz = [200, 200, 200]
        sz1 = 70
        sz = [sz1, sz1, sz1]
        img2 = pycut.zoom_to_shape(img, sz, np.uint8)
        seg2 = pycut.zoom_to_shape(seg, sz, np.uint8)
        seeds2 = pycut.zoom_to_shape(seeds, sz, np.int8)

        segparams["tile_zoom_constant"] = 0.8
        gc = pycut.ImageGraphCut(img2, segparams=segparams)
        gc.set_seeds(seeds2)
        gc.run()
        # import sed3
        # ed = sed3.sed3(gc.segmentation==0, contour=seg2)
        # ed.show()

    def test_segmentation(self):
        img, seg, seeds = self.make_data()
        gc = pycut.ImageGraphCut(img)
        gc.set_seeds(seeds)
        gc.run()
        # import sed3
        # ed = sed3.sed3(gc.segmentation==0, contour=seg)
        # ed.show()
        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            30,
        )

    def test_multiscale_indexes(self):
        # there must be some data
        img = np.zeros([32, 32, 32], dtype=np.int16)
        gc = pycut.ImageGraphCut(img)

        # mask = np.zeros([1, 4, 4], dtype=np.int16)
        # mask[0,1:3,2:] = 1
        mask = np.zeros([1, 3, 3], dtype=np.int16)
        mask[0, 1:, 1] = 1
        orig_shape = [2, 6, 6]
        # zoom = 2

        inds, mask_resized = gc._ImageGraphCut__hi2lo_multiscale_indexes(
            mask, orig_shape
        )

        expected_result = [
            [
                [0, 0, 1, 1, 2, 2],
                [0, 0, 1, 1, 2, 2],
                [3, 3, 7, 8, 4, 4],
                [3, 3, 9, 10, 4, 4],
                [5, 5, 11, 12, 6, 6],
                [5, 5, 13, 14, 6, 6],
            ],
            [
                [0, 0, 1, 1, 2, 2],
                [0, 0, 1, 1, 2, 2],
                [3, 3, 15, 16, 4, 4],
                [3, 3, 17, 18, 4, 4],
                [5, 5, 19, 20, 6, 6],
                [5, 5, 21, 22, 6, 6],
            ],
        ]

        self.assertCountEqual(inds.reshape(-1), np.array(expected_result).reshape(-1))

    def test_ordered_values_by_indexes(self):
        """
        test of pycut.__ordered_values_by_indexes
        """

        # there must be some data
        img = np.zeros([32, 32, 32], dtype=np.int16)
        data = np.array([[0, 1, 1], [0, 2, 2], [0, 2, 2]])

        inds = np.array([[0, 1, 2], [3, 4, 4], [5, 4, 4]])
        gc = pycut.ImageGraphCut(img)
        vals = gc._ImageGraphCut__ordered_values_by_indexes(data, inds)
        expected = np.array([0, 1, 1, 0, 2, 0])
        self.assertCountEqual(vals, expected)

    def test_ordered_values_by_indexes_with_different_values(self):
        """
        test of pycut.__ordered_values_by_indexes
        in input data are non-consistent data
        Function should take maximal value
        """

        # there must be some data
        img = np.zeros([32, 32, 32], dtype=np.int16)
        data = np.array([[0, 1, 1], [0, 2, 2], [0, 3, 2]])

        inds = np.array([[0, 1, 2], [3, 4, 4], [5, 4, 4]])
        gc = pycut.ImageGraphCut(img)
        vals = gc._ImageGraphCut__ordered_values_by_indexes(data, inds)
        expected = np.array([0, 1, 1, 0, 3, 0])
        self.assertCountEqual(vals, expected)

    # @unittest.skip("This test is almost done")
    def test_msgc_lo2hi(self):
        """
        Test multiscale segmentation lo2hi
        """

        img, seg, seeds = self.make_data(64, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "multiscale_graphcut_lo2hi",
            # 'method': 'multiscale_graphcut_hi2lo',
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "block_size": 8,
            "tile_zoom_constant": 1,
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams, keep_graph_properties=True)
        gc.set_seeds(seeds)
        gc.run()

        edseeds = seeds
        # gc.interactive_inspect_node()

        # node_unariesalt, node_neighboor_edges_and_weights, node_neighboor_seeds = gc.inspect_node(edseeds)

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            100,
        )

    def test_node_inspection_on_msgc_lo2hi(self):
        """
        Test multiscale segmentation lo2hi with node inspection
        """

        img, seg, seeds = self.make_data(64, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "multiscale_graphcut_lo2hi",
            # 'method': 'multiscale_graphcut_hi2lo',
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "block_size": 8,
            "tile_zoom_constant": 1,
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams, keep_graph_properties=True)
        gc.set_seeds(seeds)
        gc.run()

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            100,
        )
        # gc.interactive_inspect_node()
        node_msindex = gc.debug_get_node_msindex(seeds)

        node_unariesalt, node_neighboor_edges_and_weights, node_neighboor_seeds = gc.debug_inspect_node(
            node_msindex=node_msindex
        )
        ii, ij, ik = np.nonzero(node_neighboor_seeds)
        #
        self.assertLessEqual(
            np.max(ii) - np.min(ii),
            24,
            msg="Neighbor nodes position variance should be maximum 3 * block size ",
        )
        self.assertLessEqual(
            np.max(ij) - np.min(ij),
            24,
            msg="Neighbor nodes position variance should be maximum 3 * block size ",
        )
        self.assertLessEqual(
            np.max(ik) - np.min(ik),
            24,
            msg="Neighbor nodes position variance should be maximum 3 * block size ",
        )

    def test_msgc_lo2hi_crazy_non_block_sized_images_65(self):
        """
        Test multiscale segmentation lo2hi crazy non block sized images 65
        """

        img, seg, seeds = self.make_data(65, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "multiscale_graphcut_lo2hi",
            # 'method': 'multiscale_graphcut_hi2lo',
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "block_size": 8,
            "tile_zoom_constant": 1,
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams, keep_graph_properties=True)
        gc.set_seeds(seeds)
        gc.run()

        edseeds = seeds
        # gc.interactive_inspect_node()

        # node_unariesalt, node_neighboor_edges_and_weights, node_neighboor_seeds = gc.inspect_node(edseeds)

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            100,
        )

    def test_msgc_lo2hi_crazy_non_block_sized_images_68(self):
        """
        Test multiscale segmentation lo2hi crazy block 68
        """

        img, seg, seeds = self.make_data(68, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "multiscale_graphcut_lo2hi",
            # 'method': 'multiscale_graphcut_hi2lo',
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "block_size": 8,
            "tile_zoom_constant": 1,
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams, keep_graph_properties=True)
        gc.set_seeds(seeds)
        gc.run()

        edseeds = seeds
        # gc.interactive_inspect_node()

        # node_unariesalt, node_neighboor_edges_and_weights, node_neighboor_seeds = gc.inspect_node(edseeds)

        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            100,
        )

    def test_msgc_lo2hi_round_data(self):
        """
        Test multiscale segmentation lo2hi round data
        """
        np.random.seed(3)
        img, seg, seeds = generate_round_data(45, 3, 10, 3)
        blocksize = 6
        segparams = {
            # 'method':'graphcut',
            "method": "multiscale_graphcut_lo2hi",
            # 'method': 'multiscale_graphcut_hi2lo',
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 1,
            "boundary_penalties_weight": 1,
            "block_size": blocksize,
            "tile_zoom_constant": 1,
        }
        gc = pycut.ImageGraphCut(
            img,
            segparams=segparams,
            keep_graph_properties=True,
            # debug_images=True,
            debug_images=False,
        )
        gc.set_seeds(seeds)
        gc.run()
        # import sed3
        # ed = sed3.show_slices(img, gc.segmentation*2)

        # gc.debug_show_model()
        # gc.debug_show_reconstructed_similarity()
        tdata1, tdata2 = gc.debug_get_reconstructed_similarity()

        edseeds = seeds
        # node_unariesalt, node_neighboor_edges_and_weights, node_neighboor_seeds, node_msindex = gc.interactive_inspect_node()

        node_msindex = gc.debug_get_node_msindex(edseeds)
        node_unariesalt, node_neighboor_edges_and_weights, node_neighboor_seeds = gc.debug_inspect_node(
            node_msindex
        )

        nlinks_max = gc.debug_reconstruct_nlinks_max()

        from imcut import image_manipulation as imma

        seg_uncr = imma.uncrop(gc.segmentation, None, nlinks_max.shape)
        # import sed3
        # ed = sed3.sed3(nlinks_max, contour=seg_uncr*2)
        # ed.show()
        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            100,
        )
        self.assertEqual(
            np.max(nlinks_max),
            blocksize,
            "nlink maximal value in reconstructed nlink map. "
            "It should be size of block if there are at least some"
            " lowres blocks.",
        )
        self.assertEqual(
            np.min(nlinks_max), 1, "nlink minimal value in reconstructed nlink map"
        )
        self.assertGreaterEqual(
            np.min(tdata1), 0, "tlink minimal value in reconstructed tlink map"
        )
        self.assertGreaterEqual(
            np.min(tdata2), 0, "tlink minimal value in reconstructed tlink map"
        )
        self.assertGreaterEqual(
            node_neighboor_edges_and_weights.shape[0],
            3,
            "check number of nlink connections",
        )
        self.assertLessEqual(
            node_neighboor_edges_and_weights.shape[0],
            6,
            "check number of nlink connections",
        )
        self.assertGreaterEqual(
            np.min(node_unariesalt), 0, "selected node nlink minimum"
        )

    def test_ssgc_just_objects_with_seeds_round_data(self):
        """
        Test single scale graph cut just objects with seeds on round data
        """
        np.random.seed(3)
        img, seg, seeds = generate_round_data(
            45, 3, 10, 3, add_object_without_seeds=True
        )
        segparams = {
            "method": "graphcut",
            # 'method': 'multiscale_graphcut_hi2lo',
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "block_size": 8,
            "tile_zoom_constant": 1,
            "return_only_object_with_seeds": True,
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams, keep_graph_properties=True)
        gc.set_seeds(seeds)
        gc.run()
        # import sed3
        # ed = sed3.show_slices(img, gc.segmentation*2, seeds=seeds)

        self.assertEqual((gc.segmentation == 0).astype(np.int8)[-3, -3, -3], 0)
        self.assertLess(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            1500,
            msg="error is expected in the corner",
        )
        self.assertGreater(
            np.sum(
                np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
            ),
            1000,
            msg="error is expected in the corner",
        )

    def test_save_load(self):
        """
        test external feature vector with save model in the middle of processing
        """

        img, seg, seeds = self.make_data(64, 20)
        segparams = {
            # 'method':'graphcut',
            "method": "graphcut",
            "use_boundary_penalties": False,
            "boundary_dilatation_distance": 2,
            "boundary_penalties_weight": 1,
            "modelparams": {
                "type": "gmmsame",
                "fv_type": "intensity",
                # 'fv_extern': fv_function,
                "adaptation": "original_data",
            },
        }
        gc = pycut.ImageGraphCut(img, segparams=segparams)
        gc.set_seeds(seeds)

        gc.run()
        # import sed3
        # ed = sed3.sed3((gc.segmentation==0).astype(np.double), contour=seg)
        # ed.show()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertLess(err, 600)

        mdl_stored_file = "test_model.p"
        gc.save(mdl_stored_file)

        # forget
        gc = None

        img, seg, seeds = self.make_data(56, 18)
        # there is only one change in mdl params
        # segparams['modelparams'] = {
        #     'mdl_stored_file': mdl_stored_file,
        # }
        gc = pycut.ImageGraphCut(img)  # , segparams=segparams)
        gc.load(mdl_stored_file)
        gc.set_seeds(seeds)
        gc.run()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertLess(err, 600)
        # import sed3
        # sed3.show_slices(img, contour=gc.segmentation==0, slice_step=6)

        # if we change the data there should be more error (assertMore)
        img = (img * 0.2).astype(np.uint8)
        # segparams['modelparams']['adaptation'] = 'original_data'
        # print(np.max(img))
        # print(np.min(img))
        gc = pycut.ImageGraphCut(img)  # , segparams=segparams)
        gc.load(mdl_stored_file)
        gc.set_seeds(seeds)
        gc.run()

        m0 = gc.mdl.mdl[1]
        m1 = gc.mdl.mdl[2]
        logger.debug("model parameters")

        # import sed3
        # ed = sed3.sed3((gc.segmentation==0).astype(np.double), contour=seg)
        # ed.show()

        err = np.sum(
            np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        )
        self.assertGreater(
            err, 600, "There should be greater error when we changed the data"
        )
        # self.assertGreater(
        #     np.sum(
        #         np.abs(
        #             (gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
        #     ),
        #     600)

        os.remove(mdl_stored_file)


def generate_round_data(
    sz=32, offset=0, radius=7, seedsz=3, add_object_without_seeds=False
):
    # seedsz= int(sz/10)
    space = 2
    seeds = np.zeros([sz, sz + 1, sz + 2], dtype=np.int8)
    xmin = radius + seedsz + offset + 2
    ymin = radius + seedsz + offset + 6
    seeds[offset + 12, xmin + 3 : xmin + 7 + seedsz, ymin : ymin + 2] = 1
    seeds[offset + 20, xmin + 7 : xmin + 12 + seedsz, ymin + 5 : ymin + 7] = 1

    # add temp seed
    if add_object_without_seeds:
        seeds[-3, -3, -3] = 1
    img = np.ones([sz, sz + 1, sz + 2])
    img = img - seeds

    seeds[2 : 10 + seedsz, 2 : 9 + seedsz, 2 : 3 + seedsz] = 2

    # remove temo seed
    if add_object_without_seeds:
        seeds[-3, -3, -3] = 0

    img = scipy.ndimage.morphology.distance_transform_edt(img)
    segm = img < radius
    img = (100 * segm + 80 * np.random.random(img.shape)).astype(np.uint8)
    return img, segm, seeds


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging

logger = logging.getLogger(__name__)
import numpy as np
import scipy.ndimage



def fv_function_intensity_and_smoothing(data, voxelsize, seeds=None, unique_cls=None):
    data2 = scipy.ndimage.filters.gaussian_filter(data, sigma=5)
    # data2 = data2 - data
    arrs = [data.reshape(-1, 1), data2.reshape(-1, 1)]
    fv = np.concatenate(arrs, axis=1)
    fv2 = np.stack(arrs, -1)

    return return_fv_by_seeds(fv, seeds, unique_cls)

    # if seeds is not None:
    #     return select_from_fv_by_seeds(fv, seeds, unique_cls)
    # return fv

    # if seeds is not None:
    #     fv = []
    #     for cl in unique_cls:
    #         fv1 = data[seeds == cl].reshape(-1, 1)
    #         fv2 = data2[seeds == cl].reshape(-1, 1)
    #         fvi = np.hstack((fv1, fv2))
    #         fvi = fvi.reshape(-1, 2)
    #         fv.append(fvi)
    # else:
    #     fv1 = data.reshape(-1, 1)
    #     fv2 = data2.reshape(-1, 1)
    #     fv = np.hstack((fv1, fv2))
    #     fv = fv.reshape(-1, 2)
    #     logger.debug(str(fv[:10, :]))

def select_from_fv_by_seeds(fv, seeds, unique_cls):
    """
    Tool to make simple feature functions take features from feature array by seeds.
    :param fv: ndarray with lineariezed feature. It's shape is MxN, where M is number of image pixels and N is number
    of features
    :param seeds: ndarray with seeds. Does not to be linear.
    :param unique_cls: number of used seeds clases. Like [1, 2]
    :return: fv_selection, seeds_selection - selection from feature vector and selection from seeds
    """
    logger.debug("seeds" + str(seeds))
    # fvlin = fv.reshape(-1, int(fv.size/seeds.size))
    expected_shape = [seeds.size, int(fv.size/seeds.size)]
    if fv.shape[0] != expected_shape[0] or fv.shape[1] != expected_shape[1]:
        raise AssertionError("Wrong shape of input feature vector array fv")
    # sd = seeds.reshape(-1, 1)
    selection = np.in1d(seeds, unique_cls)
    fv_selection = fv[selection]
    seeds_selection = seeds.flatten()[selection]
    # sd = sd[]
    return fv_selection, seeds_selection

def return_fv_by_seeds(fv, seeds=None, unique_cls=None):
    """
    Return features selected by seeds and unique_cls or selection from features and corresponding seed classes.

    :param fv: ndarray with lineariezed feature. It's shape is MxN, where M is number of image pixels and N is number
    of features
    :param seeds: ndarray with seeds. Does not to be linear.
    :param unique_cls: number of used seeds clases. Like [1, 2]
    :return: fv, sd - selection from feature vector and selection from seeds or just fv for whole image
    """
    if seeds is not None:
        if unique_cls is not None:
            return select_from_fv_by_seeds(fv, seeds, unique_cls)
        else:
            raise AssertionError("Input unique_cls has to be not None if seeds is not None.")
    else:
        return fv






# cmd> snakeviz hi2lo.profile

import numpy as np
import scipy
from imcut import pycut
import cProfile
# import io3d

def make_data(sz=32, offset=0, sigma=80):
    seeds = np.zeros([sz, sz, sz], dtype=np.int8)
    seeds[offset + 12, offset + 9 : offset + 14, offset + 10] = 1
    seeds[offset + 20, offset + 18 : offset + 21, offset + 12] = 1
    img = np.ones([sz, sz, sz])
    img = img - seeds

    seeds[
    offset + 3 : offset + 15, offset + 2 : offset + 6, offset + 27 : offset + 29
    ] = 2
    img = scipy.ndimage.morphology.distance_transform_edt(img)
    segm = img < 7
    img = (100 * segm + sigma * np.random.random(img.shape)).astype(np.uint8)
    return img, segm, seeds

img, seg, seeds = make_data(64, 20)
segparams = {
    # 'method':'graphcut',
    # "method": "multiscale_graphcut",
    # "method": "hi2lo",
    "method": "lo2hi",
    "use_boundary_penalties": False,
    "boundary_dilatation_distance": 2,
    "boundary_penalties_weight": 1,
    "block_size": 8,
    "tile_zoom_constant": 1,
}
gc = pycut.ImageGraphCut(img, segparams=segparams)
gc.set_seeds(seeds)
gc.run()
# cProfile.run("gc.run()")
# import sed3
# ed = sed3.sed3(gc.segmentation==0, contour=seg)
# ed.show()

# self.assertLess(
#     np.sum(
#         np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8))
#     ),
#     600,
# )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging

logger = logging.getLogger(__name__)

import numpy as np

import os.path as op
import sklearn
import sklearn.mixture

# version comparison
from pkg_resources import parse_version
import scipy.ndimage
import scipy.stats
from . import features

if parse_version(sklearn.__version__) > parse_version("0.10"):
    # new versions
    gmm__cvtype = "covariance_type"
    gmm__cvtype_bad = "cvtype"
    defaultmodelparams = {
        "type": "gmmsame",
        "params": {"covariance_type": "full"},
        "fv_type": "intensity",
    }
else:
    gmm__cvtype = "cvtype"
    gmm__cvtype_bad = "covariance_type"
    defaultmodelparams = {
        "type": "gmmsame",
        "params": {"cvtype": "full"},
        "fv_type": "intensity",
    }
methods = ["graphcut", "multiscale_graphcut_lo2hi", "multiscale_graphcut_hi2lo"]
accepted_methods = [
    "graphcut",
    "gc",
    "multiscale_graphcut",
    "multiscale_gc",
    "msgc",
    "msgc_lo2hi",
    "lo2hi",
    "multiscale_graphcut_lo2hi",
    "msgc_hi2lo",
    "hi2lo",
    "multiscale_graphcut_hi2lo",
]


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def softplus(x, max_error=1, keep_dtype=True):
    x = np.asarray(x)
    dtype = x.dtype
    result = max_error * np.log(1 + np.exp(x / max_error))

    if keep_dtype:
        result = result.astype(dtype)
    return result


class Model3D(object):

    """ Model for image intensity. Last dimension represent feature vector.
    m = Model()
    m.train(cla, clb)
    X = numpy.random.random([2,3,4])
    # we have data 2x3 with fature vector with 4 fatures
    m.likelihood(X,0)

    modelparams['type']: type of model estimation. Gaussian mixture from EM
    algorithm is implemented as 'gmmsame'. Gaussian kernel density estimation
    is implemented as 'gaussian_kde'. General kernel estimation ('kernel')
    is from scipy version 0.14 and it is not tested.

    fv_type: feature vector type is defined with one of fallowing string
        intensity - based on seeds and data the intensity as feature vector is used
        voxel - information in voxel1 and voxel2 is used
        fv_extern - external feature vector function specified in fv_extern label
        fv001 - pixel and gaussian blur

    fv_extern:
        function `fv_function(data, voxelsize, seeds, unique_cls)`. It is used only
        if fv_type is set to "fv_extern"

    mdl_stored_file:
        string or False. Default is false. The string is path to file with stored model.
        This model is loaded and

    adaptation:
        - retrain: no adaptatin
        - original_data: train every class only once


    """

    def __init__(self, modelparams):
        # modelparams = {}
        # modelparams.update(parameters['modelparams'])
        if "params" in modelparams.keys() and gmm__cvtype_bad in modelparams["params"]:
            value = modelparams["params"].pop(gmm__cvtype_bad)
            modelparams["params"][gmm__cvtype] = value

        self.mdl = {}
        self.modelparams = defaultmodelparams.copy()
        self.modelparams.update({"adaptation": "retrain"})
        # if modelparams are updated after load, there are problems with some setting comming from outside and rewriting
        # for example "fv_type" into "intensity"
        self.modelparams.update(modelparams)
        if "mdl_stored_file" in modelparams.keys() and modelparams["mdl_stored_file"]:
            mdl_file = modelparams["mdl_stored_file"]
            self.load(mdl_file)

    def fit_from_image(self, data, voxelsize, seeds, unique_cls):
        """
        This Method allows computes feature vector and train model.

        :cls: list of index number of requested classes in seeds
        """
        fvs, clsselected = self.features_from_image(data, voxelsize, seeds, unique_cls)
        self.fit(fvs, clsselected)
        # import pdb
        # pdb.set_trace()
        # for fv, cl in zip(fvs, cls):
        #     fvs, clsselected = self.features_from_image(data, voxelsize, seeds, cl)
        #     logger.debug('cl: ' + str(cl))
        #     self.train(fv, cl)

    def save(self, filename):
        """
        Save model to pickle file. External feature function is not stored
        """
        import dill

        tmpmodelparams = self.modelparams.copy()
        # fv_extern_src = None
        fv_extern_name = None
        # try:
        #     fv_extern_src = dill.source.getsource(tmpmodelparams['fv_extern'])
        #     tmpmodelparams.pop('fv_extern')
        # except:
        #     pass

        # fv_extern_name = dill.source.getname(tmpmodelparams['fv_extern'])
        if "fv_extern" in tmpmodelparams:
            tmpmodelparams.pop("fv_extern")

        sv = {
            "modelparams": tmpmodelparams,
            "mdl": self.mdl,
            # 'fv_extern_src': fv_extern_src,
            # 'fv_extern_src_name': fv_extern_src_name,
            # 'fv_extern_name': fv_extern_src_name,
            #
        }
        sss = dill.dumps(self.modelparams)
        logger.debug("pickled " + str(sss))

        dill.dump(sv, open(filename, "wb"))

    def load(self, mdl_file):
        """
        load model from file. fv_type is not set with this function. It is expected to set it before.
        """
        import dill as pickle

        mdl_file_e = op.expanduser(mdl_file)

        sv = pickle.load(open(mdl_file_e, "rb"))
        self.mdl = sv["mdl"]
        # self.mdl[2] = self.mdl[0]
        # try:
        #     eval(sv['fv_extern_src'])
        #     eval("fv_extern_temp_name  = " + sv['fv_extern_src_name'])
        #     sv['fv_extern'] = fv_extern_temp_name
        # except:
        #     print "pomoc,necoje blbe"
        #     pass

        self.modelparams.update(sv["modelparams"])
        logger.debug("loaded model from path: " + mdl_file_e)
        # from PyQt4 import QtCore; QtCore.pyqtRemoveInputHook()
        # import ipdb; ipdb.set_trace()

    def likelihood_from_image(self, data, voxelsize, cl):
        sha = data.shape

        likel = self.likelihood(self.features_from_image(data, voxelsize), cl)
        return likel.reshape(sha)


class Model(Model3D):
    # def __init__(self, nObjects=2, modelparams={}):
    # super(Model3D, self).__init__()

    # fix change of cvtype and covariancetype
    # print modelparams

    def features_from_image(
        self, data, voxelsize, seeds=None, unique_cls=None
    ):  # , voxels=None):
        """
        Input data is 3d image

        :param data: is 3d image
        :param seeds: ndimage with same shape as data, nonzero values means seeds.
        :param unique_cls: can select only fv for seeds from specific class.
        f.e. unique_cls = [1, 2] ignores label 0

        funcion is called twice in graph cut
        first call is with all params, second is only with data.

        based on self.modelparams['fv_type'] the feature vector is computed
        keywords "intensity", "voxels", "fv001", "fv_extern"  can be used.
        modelparams['fv_type'] = 'fv_extern' allows to use external fv function

        Example of exter feature function. For easier implementation of return values use function return_fv_by_seeds().

        def fv_function(data, voxelsize, seeds=None, cl=None):
            data2 = scipy.ndimage.filters.gaussian_filter(data, sigma=5)
            arrs = [data.reshape(-1, 1), data2.reshape(-1, 1)]
            fv = np.concatenate(arrs, axis=1)
            return imcut.features.return_fv_by_seeds(fv, seeds, unique_cls)

        modelparams['fv_extern'] = fv_function
        """

        fv_type = self.modelparams["fv_type"]
        logger.debug("fv_type " + fv_type)
        fv = []
        if fv_type == "intensity":
            fv = data.reshape(-1, 1)

            if seeds is not None:
                logger.debug("seeds: %s", scipy.stats.describe(seeds, axis=None))
                sd = seeds.reshape(-1, 1)
                selection = np.in1d(sd, unique_cls)
                fv = fv[selection]
                sd = sd[selection]
                # sd = sd[]
                return fv, sd
            return fv

        # elif fv_type in ("voxels"):
        #     if seeds is not None:
        #         fv = np.asarray(voxels).reshape(-1, 1)
        #     else:
        #         fv = data
        #         fv = fv.reshape(-1, 1)
        elif fv_type in ("fv001", "FV001", "intensity_and_blur"):

            # intensity in pixel, gaussian blur intensity
            return features.fv_function_intensity_and_smoothing(
                data, voxelsize, seeds, unique_cls
            )

            # from PyQt4.QtCore import pyqtRemoveInputHook
            # pyqtRemoveInputHook()

            # print fv1.shape
            # print fv2.shape
            # print fv.shape
        elif fv_type == "fv_extern":
            fv_function = self.modelparams["fv_extern"]
            return fv_function(data, voxelsize, seeds, unique_cls)

        else:
            logger.error("Unknown feature vector type: " + self.modelparams["fv_type"])
        return fv

    # def trainFromSomething(self, data, seeds, cls, voxels):
    #     """
    #     This Method allows computes feature vector and train model.
    #
    #     :cl: scalar index number of class
    #     """
    #     for cl, voxels_i in zip(cls, voxels):
    #         logger.debug('cl: ' + str(cl))
    #         fv = self.createFV(data, seeds, cl, voxels_i)
    #         self.train(fv, cl)

    def fit(self, clx, cla):
        """

        Args:
            clx: feature vector
            cl: class, scalar or array

        Returns:

        """
        # TODO for now only sclar is used. Do not use scalar cl if future.
        # Model is not trained from other class konwledge
        # use model trained by all classes number.
        if np.isscalar(cla):
            self._fit_one_class(clx, cla)
        else:
            cla = np.asarray(cla)
            clx = np.asarray(clx)
            # import pdb
            # pdb.set_trace()
            for cli in np.unique(cla):
                selection = cla == cli
                clxsel = clx[np.nonzero(selection)[0]]
                self._fit_one_class(clxsel, cli)

    def _fit_one_class(self, clx, cl):
        """ Train clas number cl with data clx.

        Use trainFromImageAndSeeds() function if you want to use 3D image data
        as an input.

        clx: data, 2d matrix
        cl: label, integer

        label: gmmsame, gaussian_kde, dpgmm, stored
        """

        logger.debug("clx " + str(clx[:10, :]))
        logger.debug("clx type" + str(clx.dtype))
        # name = 'clx' + str(cl) + '.npy'
        # print name
        # np.save(name, clx)
        logger.debug("_fit()")
        if self.modelparams["adaptation"] == "original_data":
            if cl in self.mdl.keys():
                return
        # if True:
        #     return

        logger.debug("training continues")

        if self.modelparams["type"] == "gmmsame":
            if len(clx.shape) == 1:
                logger.warning(
                    "reshaping in train will be removed. Use \
                                \ntrainFromImageAndSeeds() function"
                )

                print("Warning deprecated feature in train() function")
                #  je to jen jednorozměrný vektor, tak je potřeba to
                # převést na 2d matici
                clx = clx.reshape(-1, 1)
            gmmparams = self.modelparams["params"]
            self.mdl[cl] = sklearn.mixture.GaussianMixture(**gmmparams)
            self.mdl[cl].fit(clx)

        elif self.modelparams["type"] == "kernel":
            # Not working (probably) in old versions of scikits
            # from sklearn.neighbors.kde import KernelDensity
            from sklearn.neighbors import KernelDensity

            # kernelmodelparams = {'kernel': 'gaussian', 'bandwidth': 0.2}
            kernelmodelparams = self.modelparams["params"]
            self.mdl[cl] = KernelDensity(**kernelmodelparams).fit(clx)
        elif self.modelparams["type"] == "gaussian_kde":
            # print clx
            import scipy.stats

            # from PyQt4.QtCore import pyqtRemoveInputHook
            # pyqtRemoveInputHook()

            # gaussian_kde works only with floating point types
            self.mdl[cl] = scipy.stats.gaussian_kde(clx.astype(np.float))
        elif self.modelparams["type"] == "dpgmm":
            # print 'clx.shape ', clx.shape
            # print 'cl ', cl
            gmmparams = self.modelparams["params"]
            self.mdl[cl] = sklearn.mixture.DPGMM(**gmmparams)
            # todo here is a hack
            # dpgmm z nějakého důvodu nefunguje pro naše data
            # vždy natrénuje jednu složku v blízkosti nuly
            # patrně to bude mít něco společného s parametrem alpha
            # přenásobí-li se to malým číslem, zázračně to chodí
            self.mdl[cl].fit(clx * 0.001)
        elif self.modelparams["type"] == "stored":
            # Classifer is trained before segmentation and stored to pickle
            import pickle

            print("stored")
            logger.warning("deprecated use of stored parameters")

            mdl_file = self.modelparams["params"]["mdl_file"]
            self.mdl = pickle.load(open(mdl_file, "rb"))

        else:
            raise NameError("Unknown model type")

            # pdb.set_trace();
            # TODO remove saving
            #         self.save('classif.p')

    def likelihood(self, x, cl):
        """
        X = numpy.random.random([2,3,4])
        # we have data 2x3 with fature vector with 4 fatures

        Use likelihoodFromImage() function for 3d image input
        m.likelihood(X,0)
        """

        # sha = x.shape
        # xr = x.reshape(-1, sha[-1])
        # outsha = sha[:-1]
        # from PyQt4.QtCore import pyqtRemoveInputHook
        # pyqtRemoveInputHook()
        logger.debug("likel " + str(x.shape))
        if self.modelparams["type"] == "gmmsame":

            px = self.mdl[cl].score_samples(x)

        # todo ošetřit více dimenzionální fv
        # px = px.reshape(outsha)
        elif self.modelparams["type"] == "kernel":
            px = self.mdl[cl].score_samples(x)
        elif self.modelparams["type"] == "gaussian_kde":
            # print x
            # np.log because it is likelihood
            # @TODO Zde je patrně problém s reshape
            # old
            # px = np.log(self.mdl[cl](x.reshape(-1)))
            # new
            px = np.log(self.mdl[cl](x))
            # px = px.reshape(outsha)
            # from PyQt4.QtCore import pyqtRemoveInputHook
            # pyqtRemoveInputHook()
        elif self.modelparams["type"] == "dpgmm":
            # todo here is a hack
            # dpgmm z nějakého důvodu nefunguje pro naše data
            # vždy natrénuje jednu složku v blízkosti nuly
            # patrně to bude mít něco společného s parametrem alpha
            # přenásobí-li se to malým číslem, zázračně to chodí
            logger.warning(".score() replaced with .score_samples() . Check it.")
            # px = self.mdl[cl].score(x * 0.01)
            px = self.mdl[cl].score_samples(x * 0.01)
        elif self.modelparams["type"] == "stored":
            px = self.mdl[cl].score(x)
        return px



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Organ segmentation

Example:

$ pycat -f head.mat -o brain.mat
"""

from __future__ import absolute_import, division, print_function
import logging

logger = logging.getLogger(__name__)

# import unittest
# from optparse import OptionParser
import argparse
import sys

# import os.path as op


from scipy.io import loadmat
import numpy as np
import time
import scipy.stats
import copy
import pygco

# from pygco import cut_from_graph

# from . import models
from .image_manipulation import (
    seed_zoom,
    zoom_to_shape,
    resize_to_shape,
    resize_to_shape_with_zoom,
    select_objects_by_seeds,
    crop,
    uncrop,
)

from .models import Model, Model3D, defaultmodelparams, methods, accepted_methods
from .graph import Graph


class ImageGraphCut:
    """
    Interactive Graph Cut.

    ImageGraphCut(data, zoom, modelparams)
    scale

    Example:

    igc = ImageGraphCut(data)
    igc.interactivity()
    igc.make_gc()
    igc.show_segmentation()

    """

    def __init__(
        self,
        img,
        modelparams={},
        segparams={},
        voxelsize=[1, 1, 1],
        debug_images=False,
        volume_unit="mm3",
        interactivity_loop_finish_fcn=None,
        keep_graph_properties=False,
    ):
        """

        Args:
            :param img: input data
            :param modelparams: parameters of model
            :param segparams: segmentation parameters
                use_apriori_if_available - set self.apriori to ndimage with same shape as img
                apriori_gamma: influence of apriory information. 0 means no influence, 1.0 is 100% use of
                apriori information
                boundary_penalties_sigma
                pairwise_alpha_per_square_unit: the pairwise_alpha is rewriten if this parameter is used.
                    pairwise_alpha = pairwise_alpha_per_square_unit / mean(voxelsize)
            :param voxelsize: size of voxel
            :param debug_images: use to show debug images with matplotlib
            :param volume_unit: define string of volume unit. Default is "mm3"
            :param keep_graph_properties: Do not delete some usefull varibales like
                msinds, unariesalt, nlinks, temp_msgc_resized_segmentation, temp_msgc_resized_img and
                temp_msgc_resized_seeds


        Returns:

        """

        logger.debug(
            "modelparams: %s, segparams: %s, voxelsize: %s, debug_images: %s",
            modelparams,
            segparams,
            voxelsize,
            debug_images,
        )

        self.voxelsize = np.asarray(voxelsize)
        if voxelsize is not None:
            self.voxel_volume = np.prod(voxelsize)
        else:
            self.voxel_volume = None

        self._update_segparams(segparams, modelparams)

        self.img = img
        self.tdata = {}
        self.segmentation = None
        # self.segparams = segparams
        self.seeds = np.zeros(self.img.shape, dtype=np.int8)
        self.debug_images = debug_images
        self.volume_unit = volume_unit



        self.interactivity_counter = 0
        self.stats = {"tlinks shape": [], "nlinks shape": []}
        self.apriori = None
        self.interactivity_loop_finish_funcion = interactivity_loop_finish_fcn
        self.keep_temp_properties = keep_graph_properties

        self.msinds = None
        self.unariesalt2 = None
        self.nlinks = None

    def _update_segparams(self, segparams={}, modelparams={}):
        # default values                              use_boundary_penalties
        # self.segparams = {'pairwiseAlpha':10, 'use_boundary_penalties':False}
        self.segparams = {
            "method": "graphcut",
            "pairwise_alpha": 20,
            "use_boundary_penalties": False,
            "boundary_penalties_sigma": 200,
            "boundary_penalties_weight": 30,
            "return_only_object_with_seeds": False,
            "use_old_similarity": True,  # New similarity not implemented @TODO
            "use_extra_features_for_training": False,
            "use_apriori_if_available": True,
            "apriori_gamma": 0.1,
        }
        if "modelparams" in segparams.keys():
            modelparams = segparams["modelparams"]
        self.segparams.update(segparams)
        if "pairwise_alpha_per_square_unit" in segparams:
            self.segparams['pairwise_alpha'] = \
                self.segparams['pairwise_alpha_per_square_unit'] / \
                np.mean(self.voxel_volume)
        self.modelparams = defaultmodelparams.copy()
        self.modelparams.update(modelparams)
        self.mdl = Model(modelparams=self.modelparams)

    def load(self, filename, fv_extern=None):
        """
        Read model stored in the file.

        :param filename: Path to file with model
        :param fv_extern: external feature vector function is passed here
        :return:
        """
        self.modelparams["mdl_stored_file"] = filename
        if fv_extern is not None:
            self.modelparams["fv_extern"] = fv_extern

        # segparams['modelparams'] = {
        #     'mdl_stored_file': mdl_stored_file,
        #     # 'fv_extern': fv_function
        # }
        self.mdl = Model(modelparams=self.modelparams)

    def interactivity_loop(self, pyed):
        # @TODO stálo by za to, přehodit tlačítka na myši. Levé má teď
        # jedničku, pravé dvojku. Pravým však zpravidla označujeme pozadí a tak
        # nám vyjde popředí jako nula a pozadí jako jednička.
        # Tím také dopadne jinak interaktivní a neinteraktivní varianta.
        # import sys
        # print "logger ", logging.getLogger().getEffectiveLevel()
        # from guppy import hpy
        # h = hpy()
        # print h.heap()
        # import pdb

        # logger.debug("obj gc   " + str(sys.getsizeof(self)))

        self.set_seeds(pyed.getSeeds())
        self.run()
        pyed.setContours(1 - self.segmentation.astype(np.int8))

        # if self.interactivity_loop_finish_funcion is None:
        #     # TODO remove this statement after lisa package update (12.5.2018)
        #     from lisa import audiosupport
        #     self.interactivity_loop_finish_funcion = audiosupport.beep

        if self.interactivity_loop_finish_funcion is not None:
            self.interactivity_loop_finish_funcion()

        self.interactivity_counter += 1
        logger.debug("interactivity counter: %s", str(self.interactivity_counter))

    def __uniform_npenalty_fcn(self, orig_shape):
        return np.ones(orig_shape, dtype=np.int8)

    def __ms_npenalty_fcn(self, axis, mask, orig_shape):
        """
        :param axis: direction of edge
        :param mask: 3d ndarray with ones where is fine resolution

        Neighboorhood penalty between small pixels should be smaller then in
        bigger tiles. This is the way how to set it.

        """
        maskz = zoom_to_shape(mask, orig_shape)

        maskz_new = np.zeros(orig_shape, dtype=np.int16)
        maskz_new[maskz == 0] = self._msgc_npenalty_table[0, axis]
        maskz_new[maskz == 1] = self._msgc_npenalty_table[1, axis]
        # import sed3
        # ed = sed3.sed3(maskz_new)
        # import ipdb; ipdb.set_trace() #  noqa BREAKPOINT

        return maskz_new

    def __msgc_step0_init(self):

        # table with size 2 * self.img.ndims
        # first axis  describe whether is the edge between lowres(0) or highres(1) voxels
        # second axis describe edge direction (edge axis)
        self._msgc_npenalty_table = np.array(
            [
                [self.segparams["block_size"] * self.segparams["tile_zoom_constant"]]
                * self.img.ndim,
                [1] * self.img.ndim,
            ]
        )
        # self.__msgc_npenalty_lowres =
        # self.__msgc_npenalty_higres = 1

    def __msgc_step12_low_resolution_segmentation(self):
        """
        Get the segmentation and the
        :return:
        """
        import scipy

        start = self._start_time
        # ===== low resolution data processing
        # default parameters
        # TODO segparams_lo and segparams_hi je tam asi zbytecně
        sparams_lo = {
            "boundary_dilatation_distance": 2,
            "block_size": 6,
            "use_boundary_penalties": True,
            "boundary_penalties_weight": 1,
            "tile_zoom_constant": 1,
        }

        sparams_lo.update(self.segparams)
        sparams_hi = copy.copy(sparams_lo)
        # sparams_lo['boundary_penalties_weight'] = (
        #         sparams_lo['boundary_penalties_weight'] *
        #         sparams_lo['block_size'])
        self.segparams = sparams_lo

        self.stats["t1"] = time.time() - start
        # step 1:  low res GC
        hiseeds = self.seeds
        # ms_zoom = 4  # 0.125 #self.segparams['scale']
        # ms_zoom = self.segparams['block_size']
        # loseeds = pyed.getSeeds()
        # logger.debug("msc " + str(np.unique(hiseeds)))
        loseeds = seed_zoom(hiseeds, self.segparams["block_size"])

        hard_constraints = True

        self.seeds = loseeds

        modelparams_hi = self.modelparams.copy()
        # feature vector will be computed from selected voxels
        self.modelparams["use_extra_features_for_training"] = True

        # TODO what with voxels? It is used from here
        # hiseeds and hiimage is used to create intensity model
        self.voxels1 = self.img[hiseeds == 1].reshape(-1, 1)
        self.voxels2 = self.img[hiseeds == 2].reshape(-1, 1)
        # this is how to compute with loseeds resolution but in wrong way
        # self.voxels1 = self.img[self.seeds == 1]
        # self.voxels2 = self.img[self.seeds == 2]

        # self.voxels1 = pyed.getSeedsVal(1)
        # self.voxels2 = pyed.getSeedsVal(2)

        img_orig = self.img

        # TODO this should be done with resize_to_shape_whith_zoom
        zoom = np.asarray(loseeds.shape).astype(np.float) / img_orig.shape
        self.img = scipy.ndimage.interpolation.zoom(img_orig, zoom, order=0)
        voxelsize_orig = self.voxelsize
        logger.debug("zoom %s", zoom)
        logger.debug("vs %s", self.voxelsize)
        self.voxelsize = self.voxelsize * zoom

        # self.img = resize_to_shape_with_zoom(img_orig, loseeds.shape, 1.0 / ms_zoom, order=0)

        # this step set the self.segmentation
        self.__single_scale_gc_run()
        # logger.debug(
        #     'segmentation - max: %d min: %d' % (
        #         np.max(self.segmentation),
        #         np.min(self.segmentation)
        #     )
        # )
        logger.debug(
            "segmentation: %s", scipy.stats.describe(self.segmentation, axis=None)
        )

        self.modelparams = modelparams_hi
        self.voxelsize = voxelsize_orig
        self.img = img_orig
        self.seeds = hiseeds
        self.stats["t2"] = time.time() - start
        return hard_constraints

    def __msgc_step3_discontinuity_localization(self):
        """
        Estimate discontinuity in basis of low resolution image segmentation.
        :return: discontinuity in low resolution
        """
        import scipy

        start = self._start_time
        seg = 1 - self.segmentation.astype(np.int8)
        self.stats["low level object voxels"] = np.sum(seg)
        self.stats["low level image voxels"] = np.prod(seg.shape)
        # in seg is now stored low resolution segmentation
        # back to normal parameters
        # step 2: discontinuity localization
        # self.segparams = sparams_hi
        seg_border = scipy.ndimage.filters.laplace(seg, mode="constant")
        logger.debug("seg_border: %s", scipy.stats.describe(seg_border, axis=None))
        # logger.debug(str(np.max(seg_border)))
        # logger.debug(str(np.min(seg_border)))
        seg_border[seg_border != 0] = 1
        logger.debug("seg_border: %s", scipy.stats.describe(seg_border, axis=None))
        # scipy.ndimage.morphology.distance_transform_edt
        boundary_dilatation_distance = self.segparams["boundary_dilatation_distance"]
        seg = scipy.ndimage.morphology.binary_dilation(
            seg_border,
            # seg,
            np.ones(
                [
                    (boundary_dilatation_distance * 2) + 1,
                    (boundary_dilatation_distance * 2) + 1,
                    (boundary_dilatation_distance * 2) + 1,
                ]
            ),
        )
        if self.keep_temp_properties:
            self.temp_msgc_lowres_discontinuity = seg
        else:
            self.temp_msgc_lowres_discontinuity = None

        if self.debug_images:
            import sed3

            pd = sed3.sed3(seg_border)  # ), contour=seg)
            pd.show()
            pd = sed3.sed3(seg)  # ), contour=seg)
            pd.show()
        # segzoom = scipy.ndimage.interpolation.zoom(seg.astype('float'), zoom,
        #                                                order=0).astype('int8')
        self.stats["t3"] = time.time() - start
        return seg

    def __msgc_step45678_hi2lo_construct_graph(self, hard_constraints, seg):
        # step 4: indexes of new dual graph

        hiseeds = self.seeds
        msinds, mask_orig = self.__hi2lo_multiscale_indexes(
            seg, self.img.shape
        )  # , ms_zoom)
        logger.debug("multiscale inds " + str(msinds.shape))
        # if deb:
        #     import sed3
        #     pd = sed3.sed3(msinds, contour=seg)
        #     pd.show()

        # intensity values for indexes
        # @TODO compute average values for low resolution
        ms_img = self.img

        # @TODO __ms_create_nlinks , use __ordered_values_by_indexes
        # import pdb; pdb.set_trace() # BREAKPOINT
        # pyed.setContours(seg)

        # there is need to set correct weights between neighbooring pixels
        # this is not nice hack.
        # @TODO reorganise segparams and create_nlinks function
        # orig_shape = img_orig.shape

        self.stats["t4"] = time.time() - self._start_time

        def local_ms_npenalty(x):
            return self.__ms_npenalty_fcn(x, seg, self.img.shape)

        # here are not unique couples of nodes
        nlinks_not_unique = self.__create_nlinks(
            ms_img,
            msinds,
            # boundary_penalties_fcn=ms_npenalty_fcn
            boundary_penalties_fcn=local_ms_npenalty,
        )
        # nlinks created
        self.stats["t5"] = time.time() - self._start_time

        # get unique set
        # remove repetitive link from one pixel to another
        nlinks = ms_remove_repetitive_link(nlinks_not_unique)
        # now remove cycle link
        self.stats["t6"] = time.time() - self._start_time
        nlinks = np.array([line for line in nlinks if line[0] != line[1]])

        self.stats["t7"] = time.time() - self._start_time
        # import ipdb; ipdb.set_trace() #  noqa BREAKPOINT
        # tlinks - indexes, data_merge
        ms_values_lin = self.__ordered_values_by_indexes(self.img, msinds)
        seeds = hiseeds
        # seeds = pyed.getSeeds()
        # if deb:
        #     import sed3
        #     se = sed3.sed3(seeds)
        #     se.show()
        ms_seeds_lin = self.__ordered_values_by_indexes(seeds, msinds)
        # logger.debug("unique seeds " + str(np.unique(seeds)))
        # logger.debug("unique seeds " + str(np.unique(ms_seeds_lin)))
        mul_mask, mul_val = self.__msgc_tlinks_area_weight_from_low_segmentation(
            mask_orig
        )
        mul_mask_lin = self.__ordered_values_by_indexes(mul_mask, msinds)
        area_weight = 1
        # TODO vyresit voxelsize
        unariesalt = self.__create_tlinks(
            ms_values_lin,
            voxelsize=self.voxelsize,
            # self.voxels1, self.voxels2,
            seeds=ms_seeds_lin,
            area_weight=area_weight,
            hard_constraints=hard_constraints,
            mul_mask=mul_mask_lin,
            mul_val=mul_val,
        )

        unariesalt2 = unariesalt.reshape(-1, 2)
        self.stats["t8"] = time.time() - self._start_time
        return nlinks, unariesalt2, msinds

    def __msgc_tlinks_area_weight_from_low_segmentation(self, loseg):
        mul_val = (self.segparams["block_size"]) ** 3
        # TODO find correct value
        # mul_val = 1
        logger.debug(
            "w: %s, loseg: %s, loseg.shape: %s",
            mul_val,
            scipy.stats.describe(loseg, axis=None),
            loseg.shape,
        )
        if loseg.shape == self.img.shape:
            loseg_resized = loseg
        else:
            # resize
            loseg_resized = zoom_to_shape(loseg, self.img.shape, dtype=np.int8)
            pass
        # area_weight = loseg_resized.astype(np.int8) * w
        mul_mask = ~loseg_resized.astype(np.bool)
        return mul_mask, mul_val

    def __msgc_step9_finish_perform_gc_and_reshape(self, nlinks, unariesalt2, msinds):
        start = self._start_time
        # create potts pairwise
        # pairwiseAlpha = -10
        pairwise = -(np.eye(2) - 1)
        pairwise = (self.segparams["pairwise_alpha"] * pairwise).astype(np.int32)

        # print 'data shape ', img_orig.shape
        # print 'nlinks sh ', nlinks.shape
        # print 'tlinks sh ', unariesalt.shape

        # print "cut_from_graph"
        # print "unaries sh ", unariesalt.reshape(-1,2).shape
        # print "nlinks sh",  nlinks.shape
        self.stats["t9"] = time.time() - start
        self.stats["tlinks shape"].append(unariesalt2.shape)
        self.stats["nlinks shape"].append(nlinks.shape)
        start_gc = time.time()
        # Same functionality is in self.seg_data()
        result_graph = pygco.cut_from_graph(
            nlinks.astype(np.int32),
            unariesalt2.astype(np.int32),
            pairwise.astype(np.int32),
        )

        elapsed = time.time() - start_gc
        self.stats["gc time"] = elapsed
        self.stats["t10"] = time.time() - start

        # probably not necessary
        #        del nlinks
        #        del unariesalt

        # print "unaries %.3g , %.3g" % (np.max(unariesalt),np.min(unariesalt))
        # @TODO get back original data
        # result_labeling = result_graph.reshape(data.shape)
        result_labeling = result_graph[msinds]
        # import py3DSeedEditor
        # ped = py3DSeedEditor.py3DSeedEditor(result_labeling)
        # ped.show()
        self.segmentation = result_labeling
        if self.keep_temp_properties:
            self.msinds = msinds
            self.unariesalt2 = unariesalt2
            self.nlinks = nlinks
        else:
            self.msinds = None
            self.unariesalt2 = None
            self.nlinks = None

    def __multiscale_gc_lo2hi_run(self):  # , pyed):
        """
        Run Graph-Cut segmentation with refinement of low resolution multiscale graph.
        In first step is performed normal GC on low resolution data
        Second step construct finer grid on edges of segmentation from first
        step.
        There is no option for use without `use_boundary_penalties`
        """
        # from PyQt4.QtCore import pyqtRemoveInputHook
        # pyqtRemoveInputHook()
        self._msgc_lo2hi_resize_init()
        self.__msgc_step0_init()

        hard_constraints = self.__msgc_step12_low_resolution_segmentation()
        # ===== high resolution data processing
        seg = self.__msgc_step3_discontinuity_localization()

        self.stats["t3.1"] = (time.time() - self._start_time)
        graph = Graph(
            seg,
            voxelsize=self.voxelsize,
            nsplit=self.segparams["block_size"],
            edge_weight_table=self._msgc_npenalty_table,
            compute_low_nodes_index=True,
        )

        # graph.run() = graph.generate_base_grid() + graph.split_voxels()
        # graph.run()
        graph.generate_base_grid()
        self.stats["t3.2"] = (time.time() - self._start_time)
        graph.split_voxels()

        self.stats["t3.3"] = (time.time() - self._start_time)

        self.stats.update(graph.stats)
        self.stats["t4"] = (time.time() - self._start_time)
        mul_mask, mul_val = self.__msgc_tlinks_area_weight_from_low_segmentation(seg)
        area_weight = 1
        unariesalt = self.__create_tlinks(
            self.img,
            self.voxelsize,
            self.seeds,
            area_weight=area_weight,
            hard_constraints=hard_constraints,
            mul_mask=None,
            mul_val=None,
        )
        # N-links prepared
        self.stats["t5"] = (time.time() - self._start_time)
        un, ind = np.unique(graph.msinds, return_index=True)
        self.stats["t6"] = (time.time() - self._start_time)

        self.stats["t7"] = (time.time() - self._start_time)
        unariesalt2_lo2hi = np.hstack(
            [unariesalt[ind, 0, 0].reshape(-1, 1), unariesalt[ind, 0, 1].reshape(-1, 1)]
        )
        nlinks_lo2hi = np.hstack([graph.edges, graph.edges_weights.reshape(-1, 1)])
        if self.debug_images:
            import sed3

            ed = sed3.sed3(unariesalt[:, :, 0].reshape(self.img.shape))
            ed.show()
            import sed3

            ed = sed3.sed3(unariesalt[:, :, 1].reshape(self.img.shape))
            ed.show()
            # ed = sed3.sed3(seg)
            # ed.show()
            # import sed3
            # ed = sed3.sed3(graph.data)
            # ed.show()
            # import sed3
            # ed = sed3.sed3(graph.msinds)
            # ed.show()

        # nlinks, unariesalt2, msinds = self.__msgc_step45678_construct_graph(area_weight, hard_constraints, seg)
        # self.__msgc_step9_finish_perform_gc_and_reshape(nlinks, unariesalt2, msinds)
        self.__msgc_step9_finish_perform_gc_and_reshape(
            nlinks_lo2hi, unariesalt2_lo2hi, graph.msinds
        )
        self._msgc_lo2hi_resize_clean_finish()

    def __multiscale_gc_hi2lo_run(self):  # , pyed):
        """
        Run Graph-Cut segmentation with simplifiyng of high resolution multiscale graph.
        In first step is performed normal GC on low resolution data
        Second step construct finer grid on edges of segmentation from first
        step.
        There is no option for use without `use_boundary_penalties`
        """
        # from PyQt4.QtCore import pyqtRemoveInputHook
        # pyqtRemoveInputHook()

        self.__msgc_step0_init()
        hard_constraints = self.__msgc_step12_low_resolution_segmentation()
        # ===== high resolution data processing
        seg = self.__msgc_step3_discontinuity_localization()
        nlinks, unariesalt2, msinds = self.__msgc_step45678_hi2lo_construct_graph(
            hard_constraints, seg
        )
        self.__msgc_step9_finish_perform_gc_and_reshape(nlinks, unariesalt2, msinds)

    def __ordered_values_by_indexes(self, data, inds):
        """
        Return values (intensities) by indexes.

        Used for multiscale graph cut.
        data = [[0 1 1],
                [0 2 2],
                [0 2 2]]

        inds = [[0 1 2],
                [3 4 4],
                [5 4 4]]

        return: [0, 1, 1, 0, 2, 0]

        If the data are not consistent, it will take the maximal value

        """
        # get unique labels and their first indexes
        # lab, linds = np.unique(inds, return_index=True)
        # compute values by indexes
        # values = data.reshape(-1)[linds]

        # alternative slow implementation
        # if there are different data on same index, it will take
        # maximal value
        # lab = np.unique(inds)
        # values = [0]*len(lab)
        # for label in lab:
        #     values[label] = np.max(data[inds == label])
        #
        # values = np.asarray(values)

        # yet another implementation
        values = [None] * (np.max(inds) + 1)

        linear_inds = inds.ravel()
        linear_data = data.ravel()
        for i in range(0, len(linear_inds)):
            # going over all data pixels

            if values[linear_inds[i]] is None:
                # this index is found for first
                values[linear_inds[i]] = linear_data[i]
            elif values[linear_inds[i]] < linear_data[i]:
                # here can be changed maximal or minimal value
                values[linear_inds[i]] = linear_data[i]

        values = np.asarray(values)

        return values

    def __hi2lo_multiscale_indexes(self, mask, orig_shape):  # , zoom):
        """
        Function computes multiscale indexes of ndarray.

        mask: Says where is original resolution (0) and where is small
        resolution (1). Mask is in small resolution.

        orig_shape: Original shape of input data.
        zoom: Usually number greater then 1

        result = [[0 1 2],
                  [3 4 4],
                  [5 4 4]]
        """

        mask_orig = zoom_to_shape(mask, orig_shape, dtype=np.int8)

        inds_small = np.arange(mask.size).reshape(mask.shape)
        inds_small_in_orig = zoom_to_shape(inds_small, orig_shape, dtype=np.int8)
        inds_orig = np.arange(np.prod(orig_shape)).reshape(orig_shape)

        # inds_orig = inds_orig * mask_orig
        inds_orig += np.max(inds_small_in_orig) + 1
        # print 'indexes'
        # import py3DSeedEditor as ped
        # import pdb; pdb.set_trace() # BREAKPOINT

        #  '==' is not the same as 'is' for numpy.array
        inds_small_in_orig[mask_orig == True] = inds_orig[mask_orig == True]  # noqa
        inds = inds_small_in_orig
        # print np.max(inds)
        # print np.min(inds)
        inds = relabel_squeeze(inds)
        logger.debug(
            "Index after relabeling: %s", scipy.stats.describe(inds, axis=None)
        )
        # logger.debug("Minimal index after relabeling: " + str(np.min(inds)))
        # inds_orig[mask_orig==True] = 0
        # inds_small_in_orig[mask_orig==False] = 0
        # inds = (inds_orig + np.max(inds_small_in_orig) + 1) + inds_small_in_orig

        return inds, mask_orig

    def interactivity(self, min_val=None, max_val=None, qt_app=None):
        """
        Interactive seed setting with 3d seed editor
        """
        from .seed_editor_qt import QTSeedEditor
        from PyQt4.QtGui import QApplication

        if min_val is None:
            min_val = np.min(self.img)

        if max_val is None:
            max_val = np.max(self.img)

        window_c = (max_val + min_val) / 2  # .astype(np.int16)
        window_w = max_val - min_val  # .astype(np.int16)

        if qt_app is None:
            qt_app = QApplication(sys.argv)

        pyed = QTSeedEditor(
            self.img,
            modeFun=self.interactivity_loop,
            voxelSize=self.voxelsize,
            seeds=self.seeds,
            volume_unit=self.volume_unit,
        )

        pyed.changeC(window_c)
        pyed.changeW(window_w)

        qt_app.exec_()

    def set_seeds(self, seeds):
        """
        Function for manual seed setting. Sets variable seeds and prepares
        voxels for density model.
        :param seeds: ndarray (0 - nothing, 1 - object, 2 - background,
        3 - object just hard constraints, no model training, 4 - background 
        just hard constraints, no model training)
        """
        if self.img.shape != seeds.shape:
            raise Exception("Seeds must be same size as input image")

        self.seeds = seeds.astype("int8")
        self.voxels1 = self.img[self.seeds == 1]
        self.voxels2 = self.img[self.seeds == 2]

    def run(self, run_fit_model=True):
        """
        Run the Graph Cut segmentation according to preset parameters.

        :param run_fit_model: Allow to skip model fit when the model is prepared before
        :return:
        """

        if run_fit_model:
            self.fit_model(self.img, self.voxelsize, self.seeds)

        self._start_time = time.time()
        if self.segparams["method"].lower() in ("graphcut", "gc"):
            self.__single_scale_gc_run()
        elif self.segparams["method"].lower() in (
            "multiscale_graphcut",
            "multiscale_gc",
            "msgc",
            "msgc_lo2hi",
            "lo2hi",
            "multiscale_graphcut_lo2hi",
        ):
            logger.debug("performing multiscale Graph-Cut lo2hi")
            self.__multiscale_gc_lo2hi_run()
        elif self.segparams["method"].lower() in (
            "msgc_hi2lo",
            "hi2lo",
            "multiscale_graphcut_hi2lo",
        ):
            logger.debug("performing multiscale Graph-Cut hi2lo")
            self.__multiscale_gc_hi2lo_run()
        else:
            logger.error("Unknown segmentation method: " + self.segparams["method"])

    def __single_scale_gc_run(self):
        res_segm = self._ssgc_prepare_data_and_run_computation(
            # self.img,
            #                      self
            # self.voxels1, self.voxels2,
            # seeds=self.seeds
        )

        if self.segparams["return_only_object_with_seeds"]:
            try:
                # because of negative problem is as 1 segmented background and
                # as 0 is segmented foreground
                # import thresholding_functions
                # newData = thresholding_functions.getPriorityObjects(
                # newData = get_priority_objects(
                #     (1 - res_segm),
                #     nObj=-1,
                #     seeds=(self.seeds == 1).nonzero(),
                #     debug=False
                # )
                newData = select_objects_by_seeds(1 - res_segm, seeds=self.seeds)
                res_segm = 1 - newData
            except:
                import traceback

                logger.warning("Cannot import thresholding_funcions")
                traceback.print_exc()

        self.segmentation = res_segm.astype(np.int8)

    def __set_hard_hard_constraints(self, tdata1, tdata2, seeds):
        """
        it works with seed labels:
        0: nothing
        1: object 1 - full seeds
        2: object 2 - full seeds
        3: object 1 - not a training seeds
        4: object 2 - not a training seeds
        """
        seeds_mask = (seeds == 1) | (seeds == 3)
        tdata2[seeds_mask] = np.max(tdata2) + 1
        tdata1[seeds_mask] = 0

        seeds_mask = (seeds == 2) | (seeds == 4)
        tdata1[seeds_mask] = np.max(tdata1) + 1
        tdata2[seeds_mask] = 0

        return tdata1, tdata2

    def _boundary_penalties_array(self, axis, sigma=None):

        import scipy.ndimage.filters as scf

        # for axis in range(0,dim):
        filtered = scf.prewitt(self.img, axis=axis)
        if sigma is None:
            sigma2 = np.var(self.img)
        else:
            sigma2 = sigma ** 2

        filtered = np.exp(-np.power(filtered, 2) / (256 * sigma2))

        # srovnán hodnot tak, aby to vycházelo mezi 0 a 100
        # cc = 10
        # filtered = ((filtered - 1)*cc) + 10
        # logger.debug(
        #     'ax %.1g max %.3g min %.3g  avg %.3g' % (
        #         axis, np.max(filtered), np.min(filtered), np.mean(filtered))
        # )
        logger.debug(
            "boundary penalties, axis: %s, filtered: %s",
            axis,
            scipy.stats.describe(filtered, axis=None),
        )
        #
        # @TODO Check why forumla with exp is not stable
        # Oproti Boykov2001b tady nedělím dvojkou. Ta je tam jen proto,
        # aby to slušně vycházelo, takže jsem si jí upravil
        # Originální vzorec je
        # Bpq = exp( - (Ip - Iq)^2 / (2 * \sigma^2) ) * 1 / dist(p,q)
        #        filtered = (-np.power(filtered,2)/(16*sigma))
        # Přičítám tu 256 což je empiricky zjištěná hodnota - aby to dobře vyšlo
        # nedávám to do exponenciely, protože je to numericky nestabilní
        # filtered = filtered + 255 # - np.min(filtered2) + 1e-30
        # Ještě by tady měl a následovat exponenciela, ale s ní je to numericky
        # nestabilní. Netuším proč.
        # if dim >= 1:
        # odecitame od sebe tentyz obrazek
        # df0 = self.img[:-1,:] - self.img[]
        # diffs.insert(0,
        return filtered

    def _reshape_unariesalt_to_similarity(self, unariesalt, shape):
        # if unariesalt is None:
        #     unariesalt = self.unariesalt2
        # if shape is None:
        #     shape = self.img.shape
        #     if hasattr(self, "temp_msgc_resized_img"):
        #         shape = self.temp_msgc_resized_img.shape
        tdata1 = unariesalt[..., 0].reshape(shape)
        tdata2 = unariesalt[..., 1].reshape(shape)
        return tdata1, tdata2

    def _debug_show_unariesalt(
        self, unariesalt, suptitle=None, slice_number=None, show=True, bins=20
    ):
        shape = self.img.shape
        # print("unariesalt dtype ", unariesalt.dtype)
        tdata1, tdata2 = self._reshape_unariesalt_to_similarity(unariesalt, shape)
        self._debug_show_tdata_images(
            tdata1,
            tdata2,
            suptitle=suptitle,
            slice_number=slice_number,
            show=show,
            bins=bins,
        )

    def _debug_show_tdata_images(
        self, tdata1, tdata2, suptitle=None, slice_number=None, show=True, bins=20
    ):
        # Show model parameters
        logger.debug("tdata1 shape %s", str(tdata1.shape))
        if slice_number is None:
            slice_number = int(tdata1.shape[0] / 2)
        try:
            import matplotlib.pyplot as plt

            fig = plt.figure()
            if suptitle is not None:
                fig.suptitle(suptitle)
            ax = fig.add_subplot(121)
            ax.imshow(tdata1[slice_number, :, :])
            # plt.colorbar(ax=ax)

            # fig = plt.figure()
            ax = fig.add_subplot(122)
            ax.imshow(tdata2[slice_number, :, :])
            # plt.colorbar(ax=ax)

            # print('tdata1 max ', np.max(tdata1), ' min ', np.min(tdata1), " dtype ", tdata1.dtype)
            # print('tdata2 max ', np.max(tdata2), ' min ', np.min(tdata2), " dtype ", tdata2.dtype)
            # logger.debug('tdata1 max ' + str(np.max(tdata1)) + ' min ' + str(np.min(tdata1)) + " dtype " +str( tdata1.dtype))
            logger.debug("tdata1: %s", scipy.stats.describe(tdata1, axis=None))
            logger.debug("tdata2: %s", scipy.stats.describe(tdata2, axis=None))
            # print('tdata2 max ', np.max(tdata2), ' min ', np.min(tdata2), " dtype ", tdata2.dtype)

            # # histogram
            # fig = plt.figure()
            # vx1 = data[seeds==1]
            # vx2 = data[seeds==2]
            # plt.hist([vx1, vx2], 30)

            # plt.hist(voxels2)

        except:
            import traceback

            print(traceback.format_exc())
            logger.debug("problem with showing debug images")

        try:
            fig = plt.figure()
            if suptitle is not None:
                fig.suptitle(suptitle)
            ax = fig.add_subplot(121)
            plt.hist(tdata1.flatten(), bins=bins)
            ax = fig.add_subplot(122)
            plt.hist(tdata2.flatten(), bins=bins)
        except:
            import traceback

            print(traceback.format_exc())

        if show:
            plt.show()
        return fig

    def debug_show_model(
        self, start=-1000, stop=1000, nsteps=400, suptitle=None, show=True
    ):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        if suptitle is not None:
            fig.suptitle(suptitle)
        ax = fig.add_subplot(111)
        hstx = np.linspace(start, stop, nsteps)
        ax.plot(hstx, np.exp(self.mdl.likelihood_from_image(hstx, self.voxelsize, 1)))
        ax.plot(hstx, np.exp(self.mdl.likelihood_from_image(hstx, self.voxelsize, 2)))
        if show:
            plt.show()
        return fig

    def fit_model(self, data=None, voxelsize=None, seeds=None):
        if data is None:
            data = self.img
        if voxelsize is None:
            voxelsize = self.voxelsize
        if seeds is None:
            seeds = self.seeds
        # TODO rewrite just for one class and call separatelly for obj and background.

        # TODO rename voxels1 and voxels2
        # voxe1s1 and voxels2 are used only in this function for multiscale graphcut
        # threre can be some

        # Dobře to fungovalo area_weight = 0.05 a cc = 6 a diference se
        # počítaly z :-1

        # self.mdl.trainFromSomething(data, seeds, 1, self.voxels1)
        # self.mdl.trainFromSomething(data, seeds, 2, self.voxels2)
        if self.segparams["use_extra_features_for_training"]:
            self.mdl.fit(self.voxels1, 1)
            self.mdl.fit(self.voxels2, 2)
        else:
            self.mdl.fit_from_image(data, voxelsize, seeds, [1, 2]),
        # as we convert to int, we need to multipy to get sensible values

    def __similarity_for_tlinks_obj_bgr(
        self,
        data,
        voxelsize,
        # voxels1, voxels2,
        # seeds, otherfeatures=None
    ):
        """
        Compute edge values for graph cut tlinks based on image intensity
        and texture.
        """
        # self.fit_model(data, voxelsize, seeds)
        # There is a need to have small vaues for good fit
        # R(obj) = -ln( Pr (Ip | O) )
        # R(bck) = -ln( Pr (Ip | B) )
        # Boykov2001b
        # ln is computed in likelihood
        tdata1 = (-(self.mdl.likelihood_from_image(data, voxelsize, 1))) * 10
        tdata2 = (-(self.mdl.likelihood_from_image(data, voxelsize, 2))) * 10

        # to spare some memory
        dtype = np.int16
        if np.any(tdata1 > 32760):
            dtype = np.float32
        if np.any(tdata2 > 32760):
            dtype = np.float32

        if self.segparams["use_apriori_if_available"] and self.apriori is not None:
            logger.debug("using apriori information")
            gamma = self.segparams["apriori_gamma"]
            a1 = (-np.log(self.apriori * 0.998 + 0.001)) * 10
            a2 = (-np.log(0.999 - (self.apriori * 0.998))) * 10
            # logger.debug('max ' + str(np.max(tdata1)) + ' min ' + str(np.min(tdata1)))
            # logger.debug('max ' + str(np.max(tdata2)) + ' min ' + str(np.min(tdata2)))
            # logger.debug('max ' + str(np.max(a1)) + ' min ' + str(np.min(a1)))
            # logger.debug('max ' + str(np.max(a2)) + ' min ' + str(np.min(a2)))
            tdata1u = (((1 - gamma) * tdata1) + (gamma * a1)).astype(dtype)
            tdata2u = (((1 - gamma) * tdata2) + (gamma * a2)).astype(dtype)
            tdata1 = tdata1u
            tdata2 = tdata2u
            # logger.debug('   max ' + str(np.max(tdata1)) + ' min ' + str(np.min(tdata1)))
            # logger.debug('   max ' + str(np.max(tdata2)) + ' min ' + str(np.min(tdata2)))
            # logger.debug('gamma ' + str(gamma))

            # import sed3
            # ed = sed3.show_slices(tdata1)
            # ed = sed3.show_slices(tdata2)
            del tdata1u
            del tdata2u
            del a1
            del a2

        # if np.any(tdata1 < 0) or np.any(tdata2 <0):
        #     logger.error("Problem with tlinks. Likelihood is < 0")

        # if self.debug_images:
        #     self.__show_debug_tdata_images(tdata1, tdata2, suptitle="likelihood")
        return tdata1, tdata2

    def __limit(self, tdata1, min_limit=0, max_error=10, max_limit=20000):
        # logger.debug('before limit max ' + np.max(tdata1), 'min ' + np.min(tdata1) + " dtype " +  tdata1.dtype)
        tdata1[tdata1 > max_limit] = max_limit
        tdata1[tdata1 < min_limit] = min_limit
        # tdata1 = models.softplus(tdata1, max_error=max_error, keep_dtype=True)
        # replace inf with large finite number
        # tdata1 = np.nan_to_num(tdata1)
        return tdata1

    def __limit_tlinks(self, tdata1, tdata2):
        tdata1 = self.__limit(tdata1)
        tdata2 = self.__limit(tdata2)

        return tdata1, tdata2

    def __create_tlinks(
        self,
        data,
        voxelsize,
        # voxels1, voxels2,
        seeds,
        area_weight,
        hard_constraints,
        mul_mask=None,
        mul_val=None,
    ):
        tdata1, tdata2 = self.__similarity_for_tlinks_obj_bgr(
            data,
            voxelsize,
            # voxels1, voxels2,
            # seeds
        )

        # logger.debug('tdata1 min %f , max %f' % (tdata1.min(), tdata1.max()))
        # logger.debug('tdata2 min %f , max %f' % (tdata2.min(), tdata2.max()))
        if hard_constraints:
            if type(seeds) == "bool":
                raise Exception(
                    "Seeds variable  not set",
                    "There is need set seed if you want use hard constraints",
                )
            tdata1, tdata2 = self.__set_hard_hard_constraints(tdata1, tdata2, seeds)

        limit = 20000
        # carefull multiplication with limit to
        if mul_mask is not None:
            divided_limit = limit / mul_val
            mm = tdata1 > divided_limit
            tdata1[mul_mask & mm] = limit
            tdata1[mul_mask & ~mm] *= mul_val
            mm = tdata2 > divided_limit
            tdata2[mul_mask & mm] = limit
            tdata2[mul_mask & ~mm] *= mul_val
        # if not np.isscalar(area_weight):
        #     area_weight = area_weight.reshape(tdata1.shape)
        tdata1 = self.__limit(tdata1)
        tdata2 = self.__limit(tdata2)
        unariesalt = (
            0
            + (
                np.dstack(
                    area_weight * [tdata1.reshape(-1, 1), tdata2.reshape(-1, 1)]
                ).copy("C")
            )
        ).astype(np.int32)
        unariesalt = self.__limit(unariesalt)
        # if self.debug_images:
        #     self.__show_debug_(unariesalt, suptitle="after weighing and limitation")
        return unariesalt

    def __create_nlinks(self, data, inds=None, boundary_penalties_fcn=None):
        """
        Compute nlinks grid from data shape information. For boundary penalties
        are data (intensities) values are used.

        ins: Default is None. Used for multiscale GC. This are indexes of
        multiscale pixels. Next example shows one superpixel witn index 2.
        inds = [
            [1 2 2],
            [3 2 2],
            [4 5 6]]

        boundary_penalties_fcn: is function with one argument - axis. It can
            it can be used for setting penalty weights between neighbooring
            pixels.

        """
        # use the gerneral graph algorithm
        # first, we construct the grid graph
        start = time.time()
        if inds is None:
            inds = np.arange(data.size).reshape(data.shape)
        # if not self.segparams['use_boundary_penalties'] and \
        #         boundary_penalties_fcn is None :
        if boundary_penalties_fcn is None:
            # This is faster for some specific format
            edgx = np.c_[inds[:, :, :-1].ravel(), inds[:, :, 1:].ravel()]
            edgy = np.c_[inds[:, :-1, :].ravel(), inds[:, 1:, :].ravel()]
            edgz = np.c_[inds[:-1, :, :].ravel(), inds[1:, :, :].ravel()]

        else:
            logger.info("use_boundary_penalties")

            bpw = self.segparams["boundary_penalties_weight"]

            bpa = boundary_penalties_fcn(2)
            # id1=inds[:, :, :-1].ravel()
            edgx = np.c_[
                inds[:, :, :-1].ravel(),
                inds[:, :, 1:].ravel(),
                # cc * np.ones(id1.shape)
                bpw * bpa[:, :, 1:].ravel(),
            ]

            bpa = boundary_penalties_fcn(1)
            # id1 =inds[:, 1:, :].ravel()
            edgy = np.c_[
                inds[:, :-1, :].ravel(),
                inds[:, 1:, :].ravel(),
                # cc * np.ones(id1.shape)]
                bpw * bpa[:, 1:, :].ravel(),
            ]

            bpa = boundary_penalties_fcn(0)
            # id1 = inds[1:, :, :].ravel()
            edgz = np.c_[
                inds[:-1, :, :].ravel(),
                inds[1:, :, :].ravel(),
                # cc * np.ones(id1.shape)]
                bpw * bpa[1:, :, :].ravel(),
            ]

        # import pdb; pdb.set_trace()
        edges = np.vstack([edgx, edgy, edgz]).astype(np.int32)
        # edges - seznam indexu hran, kteres spolu sousedi\
        elapsed = time.time() - start
        self.stats["_create_nlinks time"] = elapsed
        logger.info("__create nlinks time " + str(elapsed))
        return edges

    def debug_get_reconstructed_similarity(
        self,
        data3d=None,
        voxelsize=None,
        seeds=None,
        area_weight=1,
        hard_constraints=True,
        return_unariesalt=False,
    ):
        """
        Use actual model to calculate similarity. If no input is given the last image is used.
        :param data3d:
        :param voxelsize:
        :param seeds:
        :param area_weight:
        :param hard_constraints:
        :param return_unariesalt:
        :return:
        """
        if data3d is None:
            data3d = self.img
        if voxelsize is None:
            voxelsize = self.voxelsize
        if seeds is None:
            seeds = self.seeds

        unariesalt = self.__create_tlinks(
            data3d,
            voxelsize,
            # voxels1, voxels2,
            seeds,
            area_weight,
            hard_constraints,
        )
        if return_unariesalt:
            return unariesalt
        else:
            return self._reshape_unariesalt_to_similarity(unariesalt, data3d.shape)

    def debug_show_reconstructed_similarity(
        self,
        data3d=None,
        voxelsize=None,
        seeds=None,
        area_weight=1,
        hard_constraints=True,
        show=True,
        bins=20,
        slice_number=None,
    ):
        """
        Show tlinks.
        :param data3d: ndarray with input data
        :param voxelsize:
        :param seeds:
        :param area_weight:
        :param hard_constraints:
        :param show:
        :param bins: histogram bins number
        :param slice_number:
        :return:
        """

        unariesalt = self.debug_get_reconstructed_similarity(
            data3d,
            voxelsize=voxelsize,
            seeds=seeds,
            area_weight=area_weight,
            hard_constraints=hard_constraints,
            return_unariesalt=True,
        )

        self._debug_show_unariesalt(
            unariesalt, show=show, bins=bins, slice_number=slice_number
        )

    def debug_inspect_node(self, node_msindex):
        """
        Get info about the node. See pycut.inspect_node() for details.
        Processing is done in temporary shape.

        :param node_seed:
        :return: node_unariesalt, node_neighboor_edges_and_weights, node_neighboor_seeds
        """
        return inspect_node(self.nlinks, self.unariesalt2, self.msinds, node_msindex)

    def debug_get_node_msindex(self, node_seeds):
        """
        Get multiscale index of voxel selected by seeds.
        :param node_seeds: ndarray
        :return int with multiscale index

        """
        node_msindex = get_node_msindex(self.msinds, node_seeds)
        return node_msindex

    def debug_interactive_inspect_node(self):
        """
        Call after segmentation to see selected node neighborhood.
        User have to select one node by click.
        :return:
        """
        if (
            np.sum(
                np.abs(
                    np.asarray(self.msinds.shape) - np.asarray(self.segmentation.shape)
                )
            )
            == 0
        ):
            segmentation = self.segmentation
        else:
            segmentation = self.temp_msgc_resized_segmentation

        logger.info("Click to select one voxel of interest")
        import sed3

        ed = sed3.sed3(self.msinds, contour=segmentation == 0)
        ed.show()
        edseeds = ed.seeds
        node_msindex = get_node_msindex(self.msinds, edseeds)

        node_unariesalt, node_neighboor_edges_and_weights, node_neighboor_seeds = self.debug_inspect_node(
            node_msindex
        )
        import sed3

        ed = sed3.sed3(
            self.msinds, contour=segmentation == 0, seeds=node_neighboor_seeds
        )
        ed.show()

        return (
            node_unariesalt,
            node_neighboor_edges_and_weights,
            node_neighboor_seeds,
            node_msindex,
        )

    def debug_reconstruct_nlinks_max(self):
        """
        It will shoe the maximum weight between neighboor pixels.
        It should be different in hires area and lowres area.
        In lowres area it whould be k* higher than highres where k is tile size.
        :return:
        """
        return reconstruct_nlinks_max(self.nlinks, self.msinds)

    def _ssgc_prepare_data_and_run_computation(
        self,
        # voxels1, voxels2,
        hard_constraints=True,
        area_weight=1,
    ):
        """
        Setting of data.
        You need set seeds if you want use hard_constraints.
        """
        # from PyQt4.QtCore import pyqtRemoveInputHook
        # pyqtRemoveInputHook()
        # import pdb; pdb.set_trace() # BREAKPOINT

        unariesalt = self.__create_tlinks(
            self.img,
            self.voxelsize,
            # voxels1, voxels2,
            self.seeds,
            area_weight,
            hard_constraints,
        )
        #  některém testu  organ semgmentation dosahují unaries -15. což je podiné
        # stačí vyhodit print před if a je to vidět
        logger.debug("unaries %.3g , %.3g" % (np.max(unariesalt), np.min(unariesalt)))
        # create potts pairwise
        # pairwiseAlpha = -10
        pairwise = -(np.eye(2) - 1)
        pairwise = (self.segparams["pairwise_alpha"] * pairwise).astype(np.int32)
        # pairwise = np.array([[0,30],[30,0]]).astype(np.int32)
        # print pairwise

        self.iparams = {}

        if self.segparams["use_boundary_penalties"]:
            sigma = self.segparams["boundary_penalties_sigma"]
            # set boundary penalties function
            # Default are penalties based on intensity differences
            boundary_penalties_fcn = lambda ax: self._boundary_penalties_array(
                axis=ax, sigma=sigma
            )
        else:
            boundary_penalties_fcn = None
        nlinks = self.__create_nlinks(
            self.img, boundary_penalties_fcn=boundary_penalties_fcn
        )

        self.stats["tlinks shape"].append(unariesalt.reshape(-1, 2).shape)
        self.stats["nlinks shape"].append(nlinks.shape)
        # we flatten the unaries
        # result_graph = cut_from_graph(nlinks, unaries.reshape(-1, 2),
        # pairwise)
        start = time.time()
        if self.debug_images:
            self._debug_show_unariesalt(unariesalt)
        result_graph = pygco.cut_from_graph(nlinks, unariesalt.reshape(-1, 2), pairwise)
        elapsed = time.time() - start
        self.stats["gc time"] = elapsed
        result_labeling = result_graph.reshape(self.img.shape)

        return result_labeling

    def _msgc_lo2hi_resize_init(self):
        self._lo2hi_resize_original_shape = self.img.shape
        new_shape = (
            np.ceil(np.asarray(self.img.shape) / float(self.segparams["block_size"]))
            * self.segparams["block_size"]
        ).astype(np.int)
        crinfo = list(zip([0] * self.img.ndim, self.img.shape))
        self.img = uncrop(self.img, crinfo, new_shape, outside_mode="nearest")
        self.seeds = uncrop(self.seeds, crinfo, new_shape)

    def _msgc_lo2hi_resize_clean_finish(self):
        orig_shape = self._lo2hi_resize_original_shape
        self.temp_msgc_resized_img = self.img
        self.temp_msgc_resized_segmentation = self.segmentation
        self.temp_msgc_resized_seeds = self.seeds
        self.img = self.temp_msgc_resized_img[
            : orig_shape[0], : orig_shape[1], : orig_shape[2]
        ]
        self.segmentation = self.temp_msgc_resized_segmentation[
            : orig_shape[0], : orig_shape[1], : orig_shape[2]
        ]
        self.seeds = self.temp_msgc_resized_seeds[
            : orig_shape[0], : orig_shape[1], : orig_shape[2]
        ]
        if not self.keep_temp_properties:
            del self.temp_msgc_resized_segmentation
            del self.temp_msgc_resized_img
            del self.temp_msgc_resized_seeds
            del self._lo2hi_resize_original_shape
        # self.stats["t11"] = time.time() - self._start_time

    def save(self, filename):
        """
        Save model to file
        :param filename: Path to file
        :return:
        """
        self.mdl.save(filename)


def ms_remove_repetitive_link(nlinks_not_unique):
    # nlinks = np.array(
    #     [list(x) for x in set(tuple(x) for x in nlinks_not_unique)]
    # )
    a = nlinks_not_unique
    nlinks = (
        np.unique(a.view(np.dtype((np.void, a.dtype.itemsize * a.shape[1]))))
        .view(a.dtype)
        .reshape(-1, a.shape[1])
    )

    return nlinks


def get_neighborhood_edes(nlinks, node_msindex):
    node_neighbor_edges = np.vstack(
        [
            nlinks[np.where(nlinks[:, 0] == node_msindex)],
            nlinks[np.where(nlinks[:, 1] == node_msindex)],
        ]
    )
    return node_neighbor_edges


def inspect_node_neighborhood(nlinks, msinds, node_msindex):
    """
    Get information about one node in graph

    :param nlinks: neighboorhood edges
    :param msinds: indexes in 3d image
    :param node_msindex: int, multiscale index of selected voxel
    :return: node_neighboor_edges_and_weights, node_neighboor_seeds
    """
    # seed_indexes = np.nonzero(node_seed)
    # selected_inds = msinds[seed_indexes]
    # node_msindex = selected_inds[0]

    node_neighbor_edges = get_neighborhood_edes(nlinks, node_msindex)

    node_neighbor_seeds = np.zeros_like(msinds, dtype=np.int8)
    for neighboor_ind in np.unique(node_neighbor_edges[:, :2].ravel()):
        node_neighbor_ind = np.where(msinds == neighboor_ind)
        node_neighbor_seeds[node_neighbor_ind] = 2

    node_neighbor_seeds[np.where(msinds == node_msindex)] = 1

    # node_coordinates = np.unravel_index(selected_voxel_ind, msinds.shape)
    # node_neighbor_coordinates = np.unravel_index(np.unique(node_neighbor_edges[:, :2].ravel()), msinds.shape)
    return node_neighbor_edges, node_neighbor_seeds


def inspect_node(nlinks, unariesalt, msinds, node_msindex):
    """
    Get information about one node in graph

    :param nlinks: neighboorhood edges
    :param unariesalt: weights
    :param msinds: indexes in 3d image
    :param node_msindex: msindex of selected node. See get_node_msindex()
    :return: node_unariesalt, node_neighboor_edges_and_weights, node_neighboor_seeds
    """

    node_unariesalt = unariesalt[node_msindex]
    neigh_edges, neigh_seeds = inspect_node_neighborhood(nlinks, msinds, node_msindex)
    return node_unariesalt, neigh_edges, neigh_seeds


def get_node_msindex(msinds, node_seed):
    """
    Convert seeds-like selection of voxel to multiscale index.
    :param msinds: ndarray with indexes
    :param node_seed: ndarray with 1 where selected pixel is, or list of indexes in this array
    :return: multiscale index of first found seed
    """
    if type(node_seed) == np.ndarray:
        seed_indexes = np.nonzero(node_seed)
    elif type(node_seed) == list:
        seed_indexes = node_seed
    else:
        seed_indexes = [node_seed]

    selected_nodes_msinds = msinds[seed_indexes]
    node_msindex = selected_nodes_msinds[0]
    return node_msindex


def reconstruct_nlinks_max(nlinks, msinds):
    nlinks_max = np.zeros_like(msinds, dtype=nlinks.dtype)
    for node_msindex in np.unique(msinds):
        edges = get_neighborhood_edes(nlinks, node_msindex)
        nlinks_max[msinds == node_msindex] = np.max(edges[:, 2])
    return nlinks_max


# class Tests(unittest.TestCase):
#     def setUp(self):
#         pass

#     def test_segmentation(self):
#         data_shp = [16,16,16]
#         data = generate_data(data_shp)
#         seeds = np.zeros(data_shp)
# setting background seeds
#         seeds[:,0,0] = 1
#         seeds[6,8:-5,2] = 2
# x[4:-4, 6:-2, 1:-6] = -1

#         igc = ImageGraphCut(data)
# igc.interactivity()
# instead of interacitivity just set seeeds
#         igc.noninteractivity(seeds)

# instead of showing just test results
# igc.show_segmentation()
#         segmentation = igc.segmentation
# Testin some pixels for result
#         self.assertTrue(segmentation[0, 0, -1] == 0)
#         self.assertTrue(segmentation[7, 9, 3] == 1)
#         self.assertTrue(np.sum(segmentation) > 10)
# pdb.set_trace()
# self.assertTrue(True)


# logger.debug(igc.segmentation.shape)

usage = "%prog [options]\n" + __doc__.rstrip()
help = {
    "in_file": 'input *.mat file with "data" field',
    "out_file": "store the output matrix to the file",
    "debug": "debug mode",
    "debug_interactivity": "turn on interactive debug mode",
    "test": "run unit test",
}


def relabel_squeeze(data):
    """  Makes relabeling of data if there are unused values.  """
    palette, index = np.unique(data, return_inverse=True)
    data = index.reshape(data.shape)
    # realy slow solution
    #        unq = np.unique(data)
    #        actual_label = 0
    #        for lab in unq:
    #            data[data == lab] = actual_label
    #            actual_label += 1

    # one another solution probably slower
    # arr = data
    # data = (np.digitize(arr.reshape(-1,),np.unique(arr))-1).reshape(arr.shape)

    return data


# @profile


def main():
    # logger = logging.getLogger(__file__)
    logger = logging.getLogger()
    logger.setLevel(logging.WARNING)
    logging.basicConfig(format="%(message)s")
    ch = logging.StreamHandler()
    formatter = logging.Formatter(
        "%(levelname)-5s [%(module)s:%(funcName)s:%(lineno)d] %(message)s"
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    # parser = OptionParser(description='Organ segmentation')

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-d", "--debug", action="store_true", help=help["debug"])
    parser.add_argument(
        "-di",
        "--debug-interactivity",
        action="store_true",
        help=help["debug_interactivity"],
    )
    parser.add_argument(
        "-i", "--input-file", action="store", default=None, help=help["in_file"]
    )
    parser.add_argument("-t", "--tests", action="store_true", help=help["test"])
    parser.add_argument(
        "-o",
        "--outputfile",
        action="store",
        dest="out_filename",
        default="output.mat",
        help=help["out_file"],
    )
    # (options, args) = parser.parse_args()
    options = parser.parse_args()

    debug_images = False

    if options.debug:
        logger.setLevel(logging.DEBUG)
        # print DEBUG
        # DEBUG = True

    if options.debug_interactivity:
        debug_images = True

    # if options.tests:
    #     sys.argv[1:]=[]
    #     unittest.main()

    if options.input_file is None:
        raise IOError("No input data!")

    else:
        dataraw = loadmat(options.input_file, variable_names=["data", "voxelsize_mm"])
    # import pdb; pdb.set_trace() # BREAKPOINT

    logger.debug("\nvoxelsize_mm " + dataraw["voxelsize_mm"].__str__())

    if sys.platform == "win32":
        # hack, on windows is voxelsize read as 2D array like [[1, 0.5, 0.5]]
        dataraw["voxelsize_mm"] = dataraw["voxelsize_mm"][0]

    igc = ImageGraphCut(
        dataraw["data"],
        voxelsize=dataraw["voxelsize_mm"],
        debug_images=debug_images  # noqa
        # , modelparams={'type': 'gaussian_kde', 'params': []}
        # , modelparams={'type':'kernel', 'params':[]}  #noqa not in  old scipy
        # , modelparams={'type':'gmmsame', 'params':{'cvtype':'full', 'n_components':3}} # noqa 3 components
        # , segparams={'type': 'multiscale_gc'}  # multisc gc
        ,
        segparams={"method": "multiscale_graphcut"}  # multisc gc
        # , modelparams={'fv_type': 'fv001'}
        # , modelparams={'type': 'dpgmm', 'params': {'cvtype': 'full', 'n_components': 5, 'alpha': 10}}  # noqa 3 components
    )
    igc.interactivity()

    logger.debug("igc interactivity countr: %s", igc.interactivity_counter)

    logger.debug(igc.segmentation.shape)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import imcut.pycut
import scipy.ndimage


def make_data(sz=32, offset=0, sigma=80):
    seeds = np.zeros([sz, sz, sz], dtype=np.int8)
    seeds[offset + 12, offset + 9:offset + 14, offset + 10] = 1
    print(offset)
    seeds[offset + 20, offset + 18:offset + 21, offset + 12] = 1
    img = np.ones([sz, sz, sz])
    img = img - seeds

    seeds[
    offset + 3:offset + 15,
    offset + 2:offset + 6,
    offset + 27:offset + 29] = 2
    img = scipy.ndimage.morphology.distance_transform_edt(img)
    segm = img < 7
    img = (100 * segm + sigma * np.random.random(img.shape)).astype(np.uint8)
    return img, segm, seeds


def main():
    segparams_ssgc = {
        # 'method':'graphcut',
        'method': 'graphcut',
        'use_boundary_penalties': False,
        'boundary_dilatation_distance': 2,
        'boundary_penalties_weight': 1,
        'modelparams': {
            'type': 'gmmsame',
            "params": {
                "n_components": 2,
            },
            # "return_only_objects_with_seeds": True,
            # 'fv_type': "fv_extern",
            # 'fv_extern': fv_function,
            # 'adaptation': 'original_data',
        }
    }
    img, seg, seeds = make_data(64, 20)
    gc = imcut.pycut.ImageGraphCut(img, segparams=segparams_ssgc)
    gc.set_seeds(seeds)

    # gc.run()
    gc.interactivity()
    print(gc.segmentation)
    err = np.sum(np.abs((gc.segmentation == 0).astype(np.int8) - seg.astype(np.int8)))
    assert(err < 600)


if __name__ == "__main__":
    main()



from __future__ import absolute_import

__all__ = ['pycut',
           # 'seed_editor_qt',
           'dcmreaddata']
__version__ = "1.7.0"
# DeprecationWarning("imcut package is renamed to imcut")
# from .pycut import Model, ImageGraphCut
# from .seed_editor_qt import QTSeedEditor
# from .dcmreaddata import DicomReader
